"""
The GUI can be launched directly from the command line:

```bash
qim3d gui --iso3d
```

Or launched from a python script

```python
import qim3d

app = qim3d.gui.iso3d.Interface()
app.launch()
```

"""
import os

import gradio as gr
import numpy as np
import plotly.graph_objects as go
from scipy import ndimage

from qim3d.io import load
from qim3d.utils.logger import log

from qim3d.gui.interface import InterfaceWithExamples


#TODO img in launch should be self.img
class Interface(InterfaceWithExamples):
    def __init__(self,
                 verbose:bool = False,
                 plot_height:int = 768,
                 img = None):
        
        super().__init__(title = "Isosurfaces for 3D visualization",
                         height = 1024,
                         width = 960,
                         verbose = verbose)

        self.interface = None
        self.img = img
        self.plot_height = plot_height

    def load_data(self, gradiofile):
        try:
            self.vol = load(gradiofile.name)
            assert self.vol.ndim == 3
        except AttributeError:
            raise gr.Error("You have to select a file")
        except ValueError:
            raise gr.Error("Unsupported file format")
        except AssertionError:
            raise gr.Error(F"File has to be 3D structure. Your structure has {self.vol.ndim} dimension{'' if self.vol.ndim == 1 else 's'}")

    def resize_vol(self, display_size):
        """Resizes the loaded volume to the display size"""

        # Get original size
        original_Z, original_Y, original_X = np.shape(self.vol)
        max_size = np.max([original_Z, original_Y, original_X])
        if self.verbose:
            log.info(f"\nOriginal volume: {original_Z, original_Y, original_X}")

        # Resize for display
        self.vol = ndimage.zoom(
            input=self.vol,
            zoom = display_size / max_size,
            order=0,
            prefilter=False,
        )

        self.display_size_z, self.display_size_y, self.display_size_x = np.shape(
            self.vol
        )
        if self.verbose:
            log.info(
                f"Resized volume: {self.display_size_z, self.display_size_y, self.display_size_x}"
            )

    def save_fig(self, fig, filename):
        # Write Plotly figure to disk
        fig.write_html(filename)

    def create_fig(self, 
        gradio_file,
        display_size,
        opacity,
        opacityscale,
        only_wireframe,
        min_value,
        max_value,
        surface_count,
        colormap,
        show_colorbar,
        reversescale,
        flip_z,
        show_axis,
        show_ticks,
        show_caps,
        show_z_slice,
        slice_z_location,
        show_y_slice,
        slice_y_location,
        show_x_slice,
        slice_x_location,):

        # Load volume
        self.load_data(gradio_file)

        # Resize data for display size
        self.resize_vol(display_size)

        # Flip Z
        if flip_z:
            self.vol = np.flip(self.vol, axis=0)

        # Create 3D grid
        Z, Y, X = np.mgrid[
            0 : self.display_size_z, 0 : self.display_size_y, 0 : self.display_size_x
        ]

        if only_wireframe:
            surface_fill = 0.2
        else:
            surface_fill = 1.0

        fig = go.Figure(
            go.Volume(
                z = Z.flatten(),
                y = Y.flatten(),
                x = X.flatten(),
                value = self.vol.flatten(),
                isomin = min_value * np.max(self.vol),
                isomax = max_value * np.max(self.vol),
                cmin = np.min(self.vol),
                cmax = np.max(self.vol),
                opacity = opacity,
                opacityscale = opacityscale,
                surface_count = surface_count,
                colorscale = colormap,
                slices_z = dict(
                    show = show_z_slice,
                    locations = [int(self.display_size_z * slice_z_location)],
                ),
                slices_y = dict(
                    show = show_y_slice,
                    locations=[int(self.display_size_y * slice_y_location)],
                ),
                slices_x = dict(
                    show = show_x_slice,
                    locations = [int(self.display_size_x * slice_x_location)],
                ),
                surface = dict(fill=surface_fill),
                caps = dict(
                    x_show = show_caps,
                    y_show = show_caps,
                    z_show = show_caps,
                ),
                showscale = show_colorbar,
                colorbar=dict(
                    thickness=8, outlinecolor="#fff", len=0.5, orientation="h"
                ),
                reversescale = reversescale,
                hoverinfo = "skip",
            )
        )

        fig.update_layout(
            scene_xaxis_showticklabels = show_ticks,
            scene_yaxis_showticklabels = show_ticks,
            scene_zaxis_showticklabels = show_ticks,
            scene_xaxis_visible = show_axis,
            scene_yaxis_visible = show_axis,
            scene_zaxis_visible = show_axis,
            scene_aspectmode="data",
            height=self.plot_height,
            hovermode=False,
            scene_camera_eye=dict(x=2.0, y=-2.0, z=1.5),
        )

        filename = "iso3d.html"
        self.save_fig(fig, filename)

        return fig, filename
    
    def remove_unused_file(self):
        # Remove localthickness.tif file from working directory
        # as it otherwise is not deleted
        os.remove("iso3d.html")

    def define_interface(self, **kwargs):

        gr.Markdown(
                """
                This tool uses Plotly Volume (https://plotly.com/python/3d-volume-plots/) to create iso surfaces from voxels based on their intensity levels.
                To optimize performance when generating visualizations, set the number of voxels (_display resolution_) and isosurfaces (_total surfaces_) to lower levels. 
                """
            )

        with gr.Row():
            # Input and parameters column
            with gr.Column(scale=1, min_width=320):
                with gr.Tab("Input"):
                    # File loader
                    gradio_file = gr.File(
                        show_label=False
                    )
                with gr.Tab("Examples"):
                    gr.Examples(examples=self.img_examples, inputs=gradio_file)

                # Run button
                with gr.Row():
                    with gr.Column(scale=3, min_width=64):
                        btn_run = gr.Button(
                            value="Run 3D visualization", variant = "primary"
                        )
                    with gr.Column(scale=1, min_width=64):
                        btn_clear = gr.Button(
                            value="Clear", variant = "stop"
                        )

                with gr.Tab("Display"):
                    # Display options

                    display_size = gr.Slider(
                        32,
                        128,
                        step=4,
                        label="Display resolution",
                        info="Number of voxels for the largest dimension",
                        value=64,
                    )
                    surface_count = gr.Slider(
                        2, 16, step=1, label="Total iso-surfaces", value=6
                    )

                    show_caps = gr.Checkbox(value=False, label="Show surface caps")

                    with gr.Row():
                        opacityscale = gr.Dropdown(
                            choices=["uniform", "extremes", "min", "max"],
                            value="uniform",
                            label="Opacity scale",
                            info="Handles opacity acording to voxel value",
                        )

                        opacity = gr.Slider(
                            0.0, 1.0, step=0.1, label="Max opacity", value=0.4
                        )
                    with gr.Row():
                        min_value = gr.Slider(
                            0.0, 1.0, step=0.05, label="Min value", value=0.1
                        )
                        max_value = gr.Slider(
                            0.0, 1.0, step=0.05, label="Max value", value=1
                        )

                with gr.Tab("Slices") as slices:
                    show_z_slice = gr.Checkbox(value=False, label="Show Z slice")
                    slice_z_location = gr.Slider(
                        0.0, 1.0, step=0.05, value=0.5, label="Position"
                    )

                    show_y_slice = gr.Checkbox(value=False, label="Show Y slice")
                    slice_y_location = gr.Slider(
                        0.0, 1.0, step=0.05, value=0.5, label="Position"
                    )

                    show_x_slice = gr.Checkbox(value=False, label="Show X slice")
                    slice_x_location = gr.Slider(
                        0.0, 1.0, step=0.05, value=0.5, label="Position"
                    )

                with gr.Tab("Misc"):
                    with gr.Row():
                        colormap = gr.Dropdown(
                            choices=[
                                "Blackbody",
                                "Bluered",
                                "Blues",
                                "Cividis",
                                "Earth",
                                "Electric",
                                "Greens",
                                "Greys",
                                "Hot",
                                "Jet",
                                "Magma",
                                "Picnic",
                                "Portland",
                                "Rainbow",
                                "RdBu",
                                "Reds",
                                "Viridis",
                                "YlGnBu",
                                "YlOrRd",
                            ],
                            value="Magma",
                            label="Colormap",
                        )

                        show_colorbar = gr.Checkbox(
                            value=False, label="Show color scale"
                        )
                        reversescale = gr.Checkbox(
                            value=False, label="Reverse color scale"
                        )
                    flip_z = gr.Checkbox(value=True, label="Flip Z axis")
                    show_axis = gr.Checkbox(value=True, label="Show axis")
                    show_ticks = gr.Checkbox(value=False, label="Show ticks")
                    only_wireframe = gr.Checkbox(
                        value=False, label="Only wireframe"
                    )

                # Inputs for gradio
                inputs = [
                    gradio_file,
                    display_size,
                    opacity,
                    opacityscale,
                    only_wireframe,
                    min_value,
                    max_value,
                    surface_count,
                    colormap,
                    show_colorbar,
                    reversescale,
                    flip_z,
                    show_axis,
                    show_ticks,
                    show_caps,
                    show_z_slice,
                    slice_z_location,
                    show_y_slice,
                    slice_y_location,
                    show_x_slice,
                    slice_x_location,
                ]

            # Output display column
            with gr.Column(scale=4):
                volvizplot = gr.Plot(show_label=False)

                plot_download = gr.File(
                    interactive=False,
                    label="Download interactive plot",
                    show_label=True,
                    visible=False,
                )

            outputs = [volvizplot, plot_download]

        #####################################
        # Listeners
        #####################################

        # Clear button
        for gr_obj in outputs:
            btn_clear.click(fn=self.clear, inputs=None, outputs=gr_obj)
        # Run button
        # fmt: off
        btn_run.click(
            fn=self.create_fig, inputs = inputs, outputs = outputs).success(
            fn=self.remove_unused_file).success(
            fn=self.set_visible, inputs=None, outputs=plot_download)

if __name__ == "__main__":
    Interface().run_interface()


from .local_thickness_ import local_thickness
from .structure_tensor_ import structure_tensor
from .detection import blob_detection
from .filters import *
from .operations import *
from .cc import get_3d_cc



import gradio as gr

class QimTheme(gr.themes.Default):
    """
    QIM3D Theme for gradio interface
    The theming options are quite broad. However if there is something you can not achieve with this theme
    there is a possibility to add some more css if you override _get_css_theme function as shown at the bottom
    in comments.
    """
    def __init__(self, force_light_mode:bool = True):
        """
        Parameters:
        -----------
        - force_light_mode (bool, optional): Gradio themes have dark mode by default. 
                Qim platform is not ready for dark mode yet, thus the tools should also be in light mode.
                This sets the darkmode values to be the same as light mode values.
        """
        super().__init__()
        self.force_light_mode = force_light_mode
        self.general_values() # Not color related
        self.set_light_mode_values()
        self.set_dark_mode_values() # Checks the light mode setting inside

    def general_values(self):
        self.set_button()
        self.set_h1()
        
    def set_light_mode_values(self):
        self.set_light_primary_button()
        self.set_light_secondary_button()
        self.set_light_checkbox()
        self.set_light_cancel_button()
        self.set_light_example()

    def set_dark_mode_values(self):
        if self.force_light_mode:
            for attr in [dark_attr for dark_attr in dir(self) if not dark_attr.startswith("_") and dark_attr.endswith("dark")]:
                self.__dict__[attr] = self.__dict__[attr[:-5]] # ligth and dark attributes have same names except for '_dark' at the end
        else:
            self.set_dark_primary_button()
            # Secondary button looks good by default in dark mode
            self.set_dark_checkbox()
            self.set_dark_cancel_button()
            # Example looks good by default in dark mode

    def set_button(self):
        self.button_transition = "0.15s"
        self.button_large_text_weight = "normal"

    def set_light_primary_button(self):
        self.run_color = "#198754"
        self.button_primary_background_fill = "#FFFFFF"
        self.button_primary_background_fill_hover = self.run_color
        self.button_primary_border_color = self.run_color
        self.button_primary_text_color = self.run_color
        self.button_primary_text_color_hover = "#FFFFFF"

    def set_dark_primary_button(self):
        self.bright_run_color = "#299764"
        self.button_primary_background_fill_dark = self.button_primary_background_fill_hover
        self.button_primary_background_fill_hover_dark = self.bright_run_color
        self.button_primary_border_color_dark = self.button_primary_border_color
        self.button_primary_border_color_hover_dark = self.bright_run_color

    def set_light_secondary_button(self):
        self.button_secondary_background_fill = "white"

    def set_light_example(self):
        """
        This sets how the examples in gradio.Examples look like. Used in iso3d.
        """
        self.border_color_accent = self.neutral_100
        self.color_accent_soft = self.neutral_100

    def set_h1(self):
        self.text_xxl = "2.5rem"

    def set_light_checkbox(self):
        light_blue = "#60a5fa"
        self.checkbox_background_color_selected = light_blue
        self.checkbox_border_color_selected = light_blue
        self.checkbox_border_color_focus = light_blue

    def set_dark_checkbox(self):
        self.checkbox_border_color_dark = self.neutral_500
        self.checkbox_border_color_focus_dark = self.checkbox_border_color_focus_dark

    def set_light_cancel_button(self):
        self.cancel_color = "#dc3545"
        self.button_cancel_background_fill = "white"
        self.button_cancel_background_fill_hover = self.cancel_color
        self.button_cancel_border_color = self.cancel_color
        self.button_cancel_text_color = self.cancel_color
        self.button_cancel_text_color_hover = "white"

    def set_dark_cancel_button(self):
        self.button_cancel_background_fill_dark = self.cancel_color
        self.button_cancel_background_fill_hover_dark = "red"
        self.button_cancel_border_color_dark = self.cancel_color
        self.button_cancel_border_color_hover_dark = "red"
        self.button_cancel_text_color_dark = "white"

    # def _get_theme_css(self):
    #     sup = super()._get_theme_css()
    #     return "\n.svelte-182fdeq {\nbackground: rgba(255, 0, 0, 0.5) !important;\n}\n" + sup # You have to use !important, so it overrides other css
        


"""
!!! quote "Reference"
    Dahl, V. A., & Dahl, A. B. (2023, June). Fast Local Thickness. 2023 IEEE/CVF Conference on Computer Vision and Pattern Recognition Workshops (CVPRW). 
    <https://doi.org/10.1109/cvprw59228.2023.00456>

    ```bibtex
    @inproceedings{Dahl_2023, title={Fast Local Thickness}, 
    url={http://dx.doi.org/10.1109/CVPRW59228.2023.00456}, 
    DOI={10.1109/cvprw59228.2023.00456}, 
    booktitle={2023 IEEE/CVF Conference on Computer Vision and Pattern Recognition Workshops (CVPRW)}, 
    publisher={IEEE}, 
    author={Dahl, Vedrana Andersen and Dahl, Anders Bjorholm}, 
    year={2023}, 
    month=jun }

    ```


The GUI can be launched directly from the command line:

```bash
qim3d gui --local-thickness
```

Or launched from a python script

```python
import qim3d

app = qim3d.gui.local_thickness.Interface()
app.launch()
```

"""
import os

# matplotlib.use("Agg")
import matplotlib.pyplot as plt
import gradio as gr
import numpy as np
import tifffile
import localthickness as lt

from qim3d.io import load
from qim3d.gui.interface import InterfaceWithExamples


class Interface(InterfaceWithExamples):
    def __init__(self,
                 img = None,
                 verbose:bool = False,
                 plot_height:int = 768,
                 figsize:int = 6): 
        
        super().__init__(title = "Local thickness",
                       height = 1024,
                       width = 960,
                       verbose = verbose)

        self.plot_height = plot_height
        self.figsize = figsize
        self.img = img

    def get_result(self):
        # Get the temporary files from gradio
        temp_sets = self.interface.temp_file_sets
        for temp_set in temp_sets:
            if "localthickness" in str(temp_set):
                # Get the lsit of the temporary files
                temp_path_list = list(temp_set)

        # Files are not in creation order,
        # so we need to get find the latest
        creation_time_list = []
        for path in temp_path_list:
            creation_time_list.append(os.path.getctime(path))

        # Get index for the latest file
        file_idx = np.argmax(creation_time_list)

        # Load the temporary file
        vol_lt = load(temp_path_list[file_idx])

        return vol_lt

    def define_interface(self):
        gr.Markdown(
        "Interface for _Fast local thickness in 3D and 2D_ (https://github.com/vedranaa/local-thickness)"
        )

        with gr.Row():
            with gr.Column(scale=1, min_width=320):
                if self.img is not None:
                    data = gr.State(value=self.img)
                else:
                    with gr.Tab("Input"):
                        data = gr.File(
                            show_label=False,
                            value=self.img,
                        )
                    with gr.Tab("Examples"):
                        gr.Examples(examples=self.img_examples, inputs=data)

                with gr.Row():
                    zpos = gr.Slider(
                        minimum=0,
                        maximum=1,
                        value=0.5,
                        step=0.01,
                        label="Z position",
                        info="Local thickness is calculated in 3D, this slider controls the visualization only.",
                    )

                with gr.Tab("Parameters"):
                    gr.Markdown(
                        "It is possible to scale down the image before processing. Lower values will make the algorithm run faster, but decreases the accuracy of results."
                    )
                    lt_scale = gr.Slider(
                        0.1, 1.0, label="Scale", value=0.5, step=0.1
                    )

                    with gr.Row():
                        threshold = gr.Slider(
                            0.0,
                            1.0,
                            value=0.5,
                            step=0.05,
                            label="Threshold",
                            info="Local thickness uses a binary image, so a threshold value is needed.",
                        )

                    dark_objects = gr.Checkbox(
                        value=False,
                        label="Dark objects",
                        info="Inverts the image before thresholding. Use in case your foreground is darker than the background.",
                    )

                with gr.Tab("Display options"):
                    cmap_original = gr.Dropdown(
                        value="viridis",
                        choices=plt.colormaps(),
                        label="Colormap - input",
                        interactive=True,
                    )
                    cmap_lt = gr.Dropdown(
                        value="magma",
                        choices=plt.colormaps(),
                        label="Colormap - local thickness",
                        interactive=True,
                    )

                    nbins = gr.Slider(
                        5, 50, value=25, step=1, label="Histogram bins"
                    )

                # Run button
                with gr.Row():
                    with gr.Column(scale=3, min_width=64):
                        btn = gr.Button(
                            "Run local thickness", variant = "primary"
                        )
                    with gr.Column(scale=1, min_width=64):
                        btn_clear = gr.Button("Clear", variant = "stop")

                
            with gr.Column(scale=4):
                with gr.Row():
                    input_vol = gr.Plot(
                        show_label=True,
                        label="Original",
                        visible=True,
                    )

                    binary_vol = gr.Plot(
                        show_label=True,
                        label="Binary",
                        visible=True,
                    )

                    output_vol = gr.Plot(
                        show_label=True,
                        label="Local thickness",
                        visible=True,
                    )
                with gr.Row():
                    histogram = gr.Plot(
                        show_label=True,
                        label="Thickness histogram",
                        visible=True,
                    )
                with gr.Row():
                    lt_output = gr.File(
                        interactive=False,
                        show_label=True,
                        label="Output file",
                        visible=False,
                    )


        # Run button
        # fmt: off
        viz_input = lambda zpos, cmap: self.show_slice(self.vol, zpos, self.vmin, self.vmax, cmap)
        viz_binary = lambda zpos, cmap: self.show_slice(self.vol_binary, zpos, None, None, cmap)
        viz_output = lambda zpos, cmap: self.show_slice(self.vol_thickness, zpos, self.vmin_lt, self.vmax_lt, cmap)

        btn.click(
            fn=self.process_input, inputs = [data, dark_objects], outputs = []).success(
            fn=viz_input, inputs = [zpos, cmap_original], outputs = input_vol).success(
            fn=self.make_binary, inputs = threshold, outputs = []).success(
            fn=viz_binary, inputs = [zpos, cmap_original], outputs = binary_vol).success(
            fn=self.compute_localthickness, inputs = lt_scale, outputs = []).success(
            fn=viz_output, inputs = [zpos, cmap_lt], outputs = output_vol).success(
            fn=self.thickness_histogram, inputs = nbins, outputs = histogram).success(
            fn=self.save_lt, inputs = [], outputs = lt_output).success(
            fn=self.remove_unused_file).success(
            fn=self.set_visible, inputs= [], outputs=lt_output)

        # Clear button
        outputs = [input_vol, output_vol, binary_vol, histogram, lt_output]
        for gr_obj in outputs:
            btn_clear.click(fn=self.clear, inputs=None, outputs=gr_obj)

        btn_clear.click(fn = self.set_invisible, inputs = [], outputs = lt_output)


        # Event listeners
        zpos.change(
            fn=viz_input, inputs = [zpos, cmap_original], outputs=input_vol, show_progress=False).success(
            fn=viz_binary, inputs = [zpos, cmap_original], outputs=binary_vol, show_progress=False).success(
            fn=viz_output, inputs = [zpos, cmap_lt], outputs=output_vol, show_progress=False)
        
        cmap_original.change(
            fn=viz_input, inputs = [zpos, cmap_original],outputs=input_vol, show_progress=False).success(
            fn=viz_binary, inputs = [zpos, cmap_original], outputs=binary_vol, show_progress=False)
        
        cmap_lt.change(
            fn=viz_output, inputs = [zpos, cmap_lt], outputs=output_vol, show_progress=False
        )

        nbins.change(
            fn = self.thickness_histogram, inputs = nbins, outputs = histogram
        )
            # fmt: on

    #######################################################
    #
    #       PIPELINE
    #
    #######################################################

    def process_input(self, data, dark_objects):
        # Load volume
        try:
            self.vol = load(data.name)
            assert self.vol.ndim == 3
        except AttributeError:
            self.vol = data
        except AssertionError:
            raise gr.Error(F"File has to be 3D structure. Your structure has {self.vol.ndim} dimension{'' if self.vol.ndim == 1 else 's'}")

        if dark_objects:
            self.vol = np.invert(self.vol)

        # Get min and max values for visualization
        self.vmin = np.min(self.vol)
        self.vmax = np.max(self.vol)

    def show_slice(self, vol, zpos, vmin=None, vmax=None, cmap="viridis"):
        plt.close()
        z_idx = int(zpos * (vol.shape[0] - 1))
        fig, ax = plt.subplots(figsize=(self.figsize, self.figsize))

        ax.imshow(vol[z_idx], interpolation="nearest", cmap=cmap, vmin=vmin, vmax=vmax)

        # Adjustments
        ax.axis("off")
        fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        return fig

    def make_binary(self, threshold):
        # Make a binary volume
        # Nothing fancy, but we could add new features here
        self.vol_binary = self.vol > (threshold * np.max(self.vol))
    
    def compute_localthickness(self, lt_scale):
        self.vol_thickness = lt.local_thickness(self.vol_binary, lt_scale)

        # Valus for visualization
        self.vmin_lt = np.min(self.vol_thickness)
        self.vmax_lt = np.max(self.vol_thickness)

    def thickness_histogram(self, nbins):
        # Ignore zero thickness
        non_zero_values = self.vol_thickness[self.vol_thickness > 0]

        # Calculate histogram
        vol_hist, bin_edges = np.histogram(non_zero_values, nbins)

        fig, ax = plt.subplots(figsize=(6, 4))

        ax.bar(
            bin_edges[:-1], vol_hist, width=np.diff(bin_edges), ec="white", align="edge"
        )

        # Adjustments
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(True)
        ax.spines["bottom"].set_visible(True)
        ax.set_yscale("log")

        return fig

    def save_lt(self):
        filename = "localthickness.tif"
        # Save output image in a temp space
        tifffile.imwrite(filename, self.vol_thickness)

        return filename
    
    def remove_unused_file(self):
        # Remove localthickness.tif file from working directory
        # as it otherwise is not deleted
        os.remove('localthickness.tif')
    
if __name__ == "__main__":
    Interface().run_interface()


import matplotlib.pyplot as plt
from qim3d.utils.logger import log
import numpy as np
import ipywidgets as widgets
from IPython.display import clear_output, display
import qim3d


def circles(blobs, vol, alpha=0.5, color="#ff9900", **kwargs):
    """
    Plots the blobs found on a slice of the volume.

    This function takes in a 3D volume and a list of blobs (detected features)
    and plots the blobs on a specified slice of the volume. If no slice is specified,
    it defaults to the middle slice of the volume.

    Args:
        blobs (array-like): An array-like object of blobs, where each blob is represented
            as a 4-tuple (p, r, c, radius). Usually the result of `qim3d.processing.blob_detection(vol)`
        vol (array-like): The 3D volume on which to plot the blobs.
        alpha (float, optional): The transparency of the blobs. Defaults to 0.5.
        color (str, optional): The color of the blobs. Defaults to "#ff9900".
        **kwargs: Arbitrary keyword arguments for the `slices` function.

    Returns:
        slicer_obj (ipywidgets.interactive): An interactive widget for visualizing the blobs.

    """

    def _slicer(z_slice):
        clear_output(wait=True)
        fig = qim3d.viz.slices(
            vol,
            n_slices=1,
            position=z_slice,
            cmap="gray",
            show_position=False,
            **kwargs
        )
        # Add circles from deteced blobs
        for detected in blobs:
            z, y, x, s = detected
            if abs(z - z_slice) < s:  # The blob is in the slice

                # Adjust the radius based on the distance from the center of the sphere
                distance_from_center = abs(z - z_slice)
                angle = (
                    np.pi / 2 * (distance_from_center / s)
                )  # Angle varies from 0 at the center to pi/2 at the edge
                adjusted_radius = s * np.cos(angle)  # Radius follows a cosine curve

                if adjusted_radius > 0.5:
                    c = plt.Circle(
                        (x, y),
                        adjusted_radius,
                        color=color,
                        linewidth=0,
                        fill=True,
                        alpha=alpha,
                    )
                    fig.get_axes()[0].add_patch(c)

        display(fig)
        return fig

    position_slider = widgets.IntSlider(
        value=vol.shape[0] // 2,
        min=0,
        max=vol.shape[0] - 1,
        description="Slice",
        continuous_update=True,
    )
    slicer_obj = widgets.interactive(_slicer, z_slice=position_slider)
    slicer_obj.layout = widgets.Layout(align_items="flex-start")

    return slicer_obj



import argparse
import webbrowser
import outputformat as ouf
import qim3d

QIM_TITLE = ouf.rainbow(
    f"\n         _          _____     __ \n  ____ _(_)___ ___ |__  /____/ / \n / __ `/ / __ `__ \ /_ </ __  /  \n/ /_/ / / / / / / /__/ / /_/ /   \n\__, /_/_/ /_/ /_/____/\__,_/    \n  /_/                 v{qim3d.__version__}\n\n",
    return_str=True,
    cmap="hot",
)


def parse_tuple(arg):
    # Remove parentheses if they are included and split by comma
    return tuple(map(int, arg.strip("()").split(",")))


def main():
    parser = argparse.ArgumentParser(description="Qim3d command-line interface.")
    subparsers = parser.add_subparsers(title="Subcommands", dest="subcommand")

    # GUIs
    gui_parser = subparsers.add_parser("gui", help="Graphical User Interfaces.")

    gui_parser.add_argument(
        "--data-explorer", action="store_true", help="Run data explorer."
    )
    gui_parser.add_argument("--iso3d", action="store_true", help="Run iso3d.")
    gui_parser.add_argument(
        "--annotation-tool", action="store_true", help="Run annotation tool."
    )
    gui_parser.add_argument(
        "--local-thickness", action="store_true", help="Run local thickness tool."
    )
    gui_parser.add_argument("--host", default="0.0.0.0", help="Desired host.")
    gui_parser.add_argument(
        "--platform", action="store_true", help="Use QIM platform address"
    )
    gui_parser.add_argument(
        "--no-browser", action="store_true", help="Do not launch browser."
    )

    # Viz
    viz_parser = subparsers.add_parser("viz", help="Volumetric visualization.")
    viz_parser.add_argument("filename", default=False, help="Path to the image file")
    viz_parser.add_argument(
        "--destination", default="k3d.html", help="Path to save html file."
    )
    viz_parser.add_argument(
        "--no-browser", action="store_true", help="Do not launch browser."
    )

    # Preview
    preview_parser = subparsers.add_parser(
        "preview", help="Preview of the image in CLI"
    )
    preview_parser.add_argument(
        "filename",
        type=str,
        metavar="FILENAME",
        help="Path to image that will be displayed",
    )
    preview_parser.add_argument(
        "--slice",
        type=int,
        metavar="S",
        default=None,
        help="Specifies which slice of the image will be displayed.\nDefaults to middle slice. If number exceeds number of slices, last slice will be displayed.",
    )
    preview_parser.add_argument(
        "--axis",
        type=int,
        metavar="AX",
        default=0,
        help="Specifies from which axis will be the slice taken. Defaults to 0.",
    )
    preview_parser.add_argument(
        "--resolution",
        type=int,
        metavar="RES",
        default=80,
        help="Resolution of displayed image. Defaults to 80.",
    )
    preview_parser.add_argument(
        "--absolute_values",
        action="store_false",
        help="By default set the maximum value to be 255 so the contrast is strong. This turns it off.",
    )

    # File Convert
    convert_parser = subparsers.add_parser(
        "convert",
        help="Convert files to different formats without loading the entire file into memory",
    )
    convert_parser.add_argument(
        "input_path",
        type=str,
        metavar="Input path",
        help="Path to image that will be converted",
    )
    convert_parser.add_argument(
        "output_path",
        type=str,
        metavar="Output path",
        help="Path to save converted image",
    )
    convert_parser.add_argument(
        "--chunks",
        type=parse_tuple,
        metavar="Chunk shape",
        default=(64, 64, 64),
        help="Chunk size for the zarr file. Defaults to (64, 64, 64).",
    )
    args = parser.parse_args()

    if args.subcommand == "gui":

        arghost = args.host
        inbrowser = not args.no_browser  # Should automatically open in browser

        interface = None
        if args.data_explorer:
            interface_class = qim3d.gui.data_explorer.Interface
        elif args.iso3d:
            interface_class = qim3d.gui.iso3d.Interface
        elif args.annotation_tool:
            interface_class = qim3d.gui.annotation_tool.Interface
        elif args.local_thickness:
            interface_class = qim3d.gui.local_thickness.Interface
        else:
            print(
                "Please select a tool by choosing one of the following flags:\n\t--data-explorer\n\t--iso3d\n\t--annotation-tool\n\t--local-thickness"
            )
            return
        interface = (
            interface_class()
        )  # called here if we add another arguments to initialize

        if args.platform:
            interface.run_interface(host=arghost)
        else:
            interface.launch(inbrowser=inbrowser, force_light_mode=False)

    elif args.subcommand == "viz":
        # Load the data
        print(f"Loading data from {args.filename}")
        volume = qim3d.io.load(str(args.filename))
        print(f"Done, volume shape: {volume.shape}")

        # Make k3d plot
        print("\nGenerating k3d plot...")
        qim3d.viz.vol(volume, show=False, save=str(args.destination))
        print(f"Done, plot available at <{args.destination}>")

        if not args.no_browser:
            print("Opening in default browser...")
            webbrowser.open_new_tab(args.destination)

    elif args.subcommand == "preview":

        image = qim3d.io.load(args.filename)

        qim3d.viz.image_preview(
            image,
            image_width=args.resolution,
            axis=args.axis,
            slice=args.slice,
            relative_intensity=args.absolute_values,
        )

    elif args.subcommand == "convert":

        qim3d.io.convert(args.input_path, args.output_path, chunk_shape=args.chunks)

    elif args.subcommand is None:
        print(QIM_TITLE)
        welcome_text = (
            "\nqim3d is a Python package for 3D image processing and visualization.\n"
            f"For more information, please visit {ouf.c('https://platform.qim.dk/qim3d/', color='orange', return_str=True)}\n"
            " \n"
            "For more information on each subcommand, type 'qim3d <subcommand> --help'.\n"
        )
        print(welcome_text)
        parser.print_help()
        print("\n")


if __name__ == "__main__":
    main()



"""Provides filter functions and classes for image processing"""

from typing import Type, Union

import numpy as np
from scipy import ndimage
from skimage import morphology

from qim3d.utils.logger import log

__all__ = [
    "Gaussian",
    "Median",
    "Maximum",
    "Minimum",
    "Pipeline",
    "Tophat",
    "gaussian",
    "median",
    "maximum",
    "minimum",
    "tophat",
]


class FilterBase:
    def __init__(self, *args, **kwargs):
        """
        Base class for image filters.

        Args:
            *args: Additional positional arguments for filter initialization.
            **kwargs: Additional keyword arguments for filter initialization.
        """
        self.args = args
        self.kwargs = kwargs


class Gaussian(FilterBase):
    def __call__(self, input):
        """
        Applies a Gaussian filter to the input.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume.
        """
        return gaussian(input, *self.args, **self.kwargs)


class Median(FilterBase):
    def __call__(self, input):
        """
        Applies a median filter to the input.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume.
        """
        return median(input, **self.kwargs)


class Maximum(FilterBase):
    def __call__(self, input):
        """
        Applies a maximum filter to the input.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume.
        """
        return maximum(input, **self.kwargs)


class Minimum(FilterBase):
    def __call__(self, input):
        """
        Applies a minimum filter to the input.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume.
        """
        return minimum(input, **self.kwargs)

class Tophat(FilterBase):
    def __call__(self, input):
        """
        Applies a tophat filter to the input.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume.
        """
        return tophat(input, **self.kwargs)


class Pipeline:
    """
    Example:
        ```python
        import qim3d
        from qim3d.processing import Pipeline, Median, Gaussian, Maximum, Minimum

        # Get data
        vol = qim3d.examples.fly_150x256x256

        # Show original
        qim3d.viz.slices(vol, axis=0, show=True)

        # Create filter pipeline
        pipeline = Pipeline(
            Median(size=5),
            Gaussian(sigma=3)
        )

        # Append a third filter to the pipeline
        pipeline.append(Maximum(size=3))

        # Apply filter pipeline
        vol_filtered = pipeline(vol)

        # Show filtered
        qim3d.viz.slices(vol_filtered, axis=0)
        ```
        ![original volume](assets/screenshots/filter_original.png)
        ![filtered volume](assets/screenshots/filter_processed.png)
            
        """
    def __init__(self, *args: Type[FilterBase]):
        """
        Represents a sequence of image filters.

        Args:
            *args: Variable number of filter instances to be applied sequentially.
        """
        self.filters = {}

        for idx, fn in enumerate(args):
            self._add_filter(str(idx), fn)

    def _add_filter(self, name: str, fn: Type[FilterBase]):
        """
        Adds a filter to the sequence.

        Args:
            name: A string representing the name or identifier of the filter.
            fn: An instance of a FilterBase subclass.

        Raises:
            AssertionError: If `fn` is not an instance of the FilterBase class.
        """
        if not isinstance(fn, FilterBase):
            filter_names = [
                subclass.__name__ for subclass in FilterBase.__subclasses__()
            ]
            raise AssertionError(
                f"filters should be instances of one of the following classes: {filter_names}"
            )
        self.filters[name] = fn

    def append(self, fn: Type[FilterBase]):
        """
        Appends a filter to the end of the sequence.

        Args:
            fn: An instance of a FilterBase subclass to be appended.
        
        Example:
            ```python
            import qim3d
            from qim3d.processing import Pipeline, Maximum, Median

            # Create filter pipeline
            pipeline = Pipeline(
                Maximum(size=3)
            )

            # Append a second filter to the pipeline
            pipeline.append(Median(size=5))
            ```
        """
        self._add_filter(str(len(self.filters)), fn)

    def __call__(self, input):
        """
        Applies the sequential filters to the input in order.

        Args:
            input: The input image or volume.

        Returns:
            The filtered image or volume after applying all sequential filters.
        """
        for fn in self.filters.values():
            input = fn(input)
        return input


def gaussian(vol, *args, **kwargs):
    """
    Applies a Gaussian filter to the input volume using scipy.ndimage.gaussian_filter.

    Args:
        vol: The input image or volume.
        *args: Additional positional arguments for the Gaussian filter.
        **kwargs: Additional keyword arguments for the Gaussian filter.

    Returns:
        The filtered image or volume.
    """
    return ndimage.gaussian_filter(vol, *args, **kwargs)


def median(vol, **kwargs):
    """
    Applies a median filter to the input volume using scipy.ndimage.median_filter.

    Args:
        vol: The input image or volume.
        **kwargs: Additional keyword arguments for the median filter.

    Returns:
        The filtered image or volume.
    """
    return ndimage.median_filter(vol, **kwargs)


def maximum(vol, **kwargs):
    """
    Applies a maximum filter to the input volume using scipy.ndimage.maximum_filter.

    Args:
        vol: The input image or volume.
        **kwargs: Additional keyword arguments for the maximum filter.

    Returns:
        The filtered image or volume.
    """
    return ndimage.maximum_filter(vol, **kwargs)


def minimum(vol, **kwargs):
    """
    Applies a minimum filter to the input volume using scipy.ndimage.mainimum_filter.

    Args:
        vol: The input image or volume.
        **kwargs: Additional keyword arguments for the minimum filter.

    Returns:
        The filtered image or volume.
    """
    return ndimage.minimum_filter(vol, **kwargs)

def tophat(vol, **kwargs):
    """
    Remove background from the volume.

    Args:
        vol: The volume to remove background from
        radius: The radius of the structuring element (default: 3)
        background: color of the background, 'dark' or 'bright' (default: 'dark'). If 'bright', volume will be inverted.
    
    Returns:
        vol: The volume with background removed
    """
    radius = kwargs["radius"] if "radius" in kwargs else 3
    background = kwargs["background"] if "background" in kwargs else "dark"
    
    if background == "bright":
            log.info("Bright background selected, volume will be temporarily inverted when applying white_tophat")
            vol = np.invert(vol)
    
    selem = morphology.ball(radius)
    vol = vol - morphology.white_tophat(vol, selem)

    if background == "bright":
        vol = np.invert(vol)
    
    return vol



"""Visualization tools"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib import colormaps
from qim3d.utils.logger import log


def plot_metrics(
    *metrics,
    linestyle="-",
    batch_linestyle="dotted",
    labels: list = None,
    figsize: tuple = (16, 6),
    show=False
):
    """
    Plots the metrics over epochs and batches.

    Args:
        *metrics: Variable-length argument list of dictionary containing the metrics per epochs and per batches.
        linestyle (str, optional): The style of the epoch metric line. Defaults to '-'.
        batch_linestyle (str, optional): The style of the batch metric line. Defaults to 'dotted'.
        labels (list[str], optional): Labels for the plotted lines. Defaults to None.
        figsize (Tuple[int, int], optional): Figure size (width, height) in inches. Defaults to (16, 6).
        show (bool, optional): If True, displays the plot. Defaults to False.

    Returns:
        fig (matplotlib.figure.Figure): plot with metrics.

    Example:
        train_loss = {'epoch_loss' : [...], 'batch_loss': [...]}
        val_loss = {'epoch_loss' : [...], 'batch_loss': [...]}
        plot_metrics(train_loss,val_loss, labels=['Train','Valid.'])
    """
    import seaborn as snb

    if labels == None:
        labels = [None] * len(metrics)
    elif len(metrics) != len(labels):
        raise ValueError("The number of metrics doesn't match the number of labels.")

    # plotting parameters
    snb.set_style("darkgrid")
    snb.set(font_scale=1.5)
    plt.rcParams["lines.linewidth"] = 2

    fig = plt.figure(figsize=figsize)

    palette = snb.color_palette(None, len(metrics))

    for i, metric in enumerate(metrics):
        metric_name = list(metric.keys())[0]
        epoch_metric = metric[list(metric.keys())[0]]
        batch_metric = metric[list(metric.keys())[1]]

        x_axis = np.linspace(0, len(epoch_metric) - 1, len(batch_metric))

        plt.plot(epoch_metric, linestyle=linestyle, color=palette[i], label=labels[i])
        plt.plot(
            x_axis, batch_metric, linestyle=batch_linestyle, color=palette[i], alpha=0.4
        )

    if labels[0] != None:
        plt.legend()

    plt.ylabel(metric_name)
    plt.xlabel("epoch")

    # reset plotting parameters
    snb.set_style("white")

    if show:
        plt.show()
    plt.close()

    return fig


def grid_overview(
    data, num_images=7, cmap_im="gray", cmap_segm="viridis", alpha=0.5, show=False
):
    """Displays an overview grid of images, labels, and masks (if they exist).

    Labels are the annotated target segmentations
    Masks are applied to the output and target prior to the loss calculation in case of
    sparse labeled data

    Args:
        data (list or torch.utils.data.Dataset): A list of tuples or Torch dataset containing image, label, (and mask data).
        num_images (int, optional): The maximum number of images to display. Defaults to 7.
        cmap_im (str, optional): The colormap to be used for displaying input images. Defaults to 'gray'.
        cmap_segm (str, optional): The colormap to be used for displaying labels. Defaults to 'viridis'.
        alpha (float, optional): The transparency level of the label and mask overlays. Defaults to 0.5.
        show (bool, optional): If True, displays the plot (i.e. calls plt.show()). Defaults to False.

    Raises:
        ValueError: If the data elements are not tuples.


    Returns:
        fig (matplotlib.figure.Figure): The figure with an overview of the images and their labels.

    Example:
        ```python
        data = [(image1, label1, mask1), (image2, label2, mask2)]
        grid_overview(data, num_images=5, cmap_im='viridis', cmap_segm='hot', alpha=0.8)
        ```

    Notes:
        - If the image data is RGB, the color map is ignored and the user is informed.
        - The number of displayed images is limited to the minimum between `num_images`
            and the length of the data.
        - The grid layout and dimensions vary based on the presence of a mask.
    """

    # Check if data has a mask
    has_mask = len(data[0]) > 2 and data[0][-1] is not None

    # Check if image data is RGB and inform the user if it's the case
    if len(data[0][0].squeeze().shape) > 2:
        log.info("Input images are RGB: color map is ignored")

    # Check if dataset have at least specified number of images
    if len(data) < num_images:
        log.warning(
            "Not enough images in the dataset. Changing num_images=%d to num_images=%d",
            num_images,
            len(data),
        )
        num_images = len(data)

    # Adapt segmentation cmap so that background is transparent
    colors_segm = colormaps.get_cmap(cmap_segm)(np.linspace(0, 1, 256))
    colors_segm[:128, 3] = 0
    custom_cmap = LinearSegmentedColormap.from_list("CustomCmap", colors_segm)

    # Check if data have the right format
    if not isinstance(data[0], tuple):
        raise ValueError("Data elements must be tuples")

    # Define row titles
    row_titles = ["Input images", "Ground truth segmentation", "Mask"]

    # Make new list such that possible augmentations remain identical for all three rows
    plot_data = [data[idx] for idx in range(num_images)]

    fig = plt.figure(
        figsize=(2 * num_images, 9 if has_mask else 6), constrained_layout=True
    )

    # create 2 (3) x 1 subfigs
    subfigs = fig.subfigures(nrows=3 if has_mask else 2, ncols=1)
    for row, subfig in enumerate(subfigs):
        subfig.suptitle(row_titles[row], fontsize=22)

        # create 1 x num_images subplots per subfig
        axs = subfig.subplots(nrows=1, ncols=num_images)
        for col, ax in enumerate(np.atleast_1d(axs)):
            if row in [1, 2]:  # Ground truth segmentation and mask
                ax.imshow(plot_data[col][0].squeeze(), cmap=cmap_im)
                ax.imshow(plot_data[col][row].squeeze(), cmap=custom_cmap, alpha=alpha)
                ax.axis("off")
            else:
                ax.imshow(plot_data[col][row].squeeze(), cmap=cmap_im)
                ax.axis("off")

    if show:
        plt.show()
    plt.close()

    return fig


def grid_pred(
    in_targ_preds,
    num_images=7,
    cmap_im="gray",
    cmap_segm="viridis",
    alpha=0.5,
    show=False,
):
    """Displays a grid of input images, predicted segmentations, ground truth segmentations, and their comparison.

    Displays a grid of subplots representing different aspects of the input images and segmentations.
    The grid includes the following rows:
        - Row 1: Input images
        - Row 2: Predicted segmentations overlaying input images
        - Row 3: Ground truth segmentations overlaying input images
        - Row 4: Comparison between true and predicted segmentations overlaying input images

    Each row consists of `num_images` subplots, where each subplot corresponds to an image from the dataset.
    The function utilizes various color maps for visualization and applies transparency to the segmentations.

    Args:
        in_targ_preds (tuple): A tuple containing input images, target segmentations, and predicted segmentations.
        num_images (int, optional): Number of images to display. Defaults to 7.
        cmap_im (str, optional): Color map for input images. Defaults to "gray".
        cmap_segm (str, optional): Color map for segmentations. Defaults to "viridis".
        alpha (float, optional): Alpha value for transparency. Defaults to 0.5.
        show (bool, optional): If True, displays the plot (i.e. calls plt.show()). Defaults to False.

    Returns:
        fig (matplotlib.figure.Figure): The figure with images, labels and the label prediction from the trained models.

    Raises:
        None

    Example:
        dataset = MySegmentationDataset()
        model = MySegmentationModel()
        in_targ_preds = qim3d.utils.models.inference(dataset,model)
        grid_pred(in_targ_preds, cmap_im='viridis', alpha=0.5)
    """
    import torch

    # Check if dataset have at least specified number of images
    if len(in_targ_preds[0]) < num_images:
        log.warning(
            "Not enough images in the dataset. Changing num_images=%d to num_images=%d",
            num_images,
            len(in_targ_preds[0]),
        )
        num_images = len(in_targ_preds[0])

    # Take only the number of images from in_targ_preds
    inputs, targets, preds = [items[:num_images] for items in in_targ_preds]

    # Adapt segmentation cmap so that background is transparent
    colors_segm = colormaps.get_cmap(cmap_segm)(np.linspace(0, 1, 256))
    colors_segm[:128, 3] = 0
    custom_cmap = LinearSegmentedColormap.from_list("CustomCmap", colors_segm)

    N = num_images
    H = inputs[0].shape[-2]
    W = inputs[0].shape[-1]

    comp_rgb = torch.zeros((N, 4, H, W))
    comp_rgb[:, 1, :, :] = targets.logical_and(preds)
    comp_rgb[:, 0, :, :] = targets.logical_xor(preds)
    comp_rgb[:, 3, :, :] = targets.logical_or(preds)

    row_titles = [
        "Input images",
        "Predicted segmentation",
        "Ground truth segmentation",
        "True vs. predicted segmentation",
    ]

    fig = plt.figure(figsize=(2 * num_images, 10), constrained_layout=True)

    # create 3 x 1 subfigs
    subfigs = fig.subfigures(nrows=4, ncols=1)
    for row, subfig in enumerate(subfigs):
        subfig.suptitle(row_titles[row], fontsize=22)

        # create 1 x num_images subplots per subfig
        axs = subfig.subplots(nrows=1, ncols=num_images)
        for col, ax in enumerate(np.atleast_1d(axs)):
            if row == 0:
                ax.imshow(inputs[col], cmap=cmap_im)
                ax.axis("off")

            elif row == 1:  # Predicted segmentation
                ax.imshow(inputs[col], cmap=cmap_im)
                ax.imshow(preds[col], cmap=custom_cmap, alpha=alpha)
                ax.axis("off")
            elif row == 2:  # Ground truth segmentation
                ax.imshow(inputs[col], cmap=cmap_im)
                ax.imshow(targets[col], cmap=custom_cmap, alpha=alpha)
                ax.axis("off")
            else:
                ax.imshow(inputs[col], cmap=cmap_im)
                ax.imshow(comp_rgb[col].permute(1, 2, 0), alpha=alpha)
                ax.axis("off")

    if show:
        plt.show()
    plt.close()

    return fig


def vol_masked(vol, vol_mask, viz_delta=128):
    """
    Applies masking to a volume based on a binary volume mask.

    This function takes a volume array `vol` and a corresponding binary volume mask `vol_mask`.
    It computes the masked volume where pixels outside the mask are set to the background value,
    and pixels inside the mask are set to foreground.


    Args:
        vol (ndarray): The input volume as a NumPy array.
        vol_mask (ndarray): The binary mask volume as a NumPy array with the same shape as `vol`.
        viz_delta (int, optional): Value added to the volume before applying the mask to visualize masked regions.
            Defaults to 128.

    Returns:
        ndarray: The masked volume with the same shape as `vol`, where pixels outside the mask are set
            to the background value (negative).


    """

    background = (vol.astype("float") + viz_delta) * (1 - vol_mask) * -1
    foreground = (vol.astype("float") + viz_delta) * vol_mask
    vol_masked_result = background + foreground

    return vol_masked_result



"""Class for choosing the level of data augmentations with albumentations"""

class Augmentation:
    """
    Class for defining image augmentation transformations using the Albumentations library.
        
    Args:
        resize (str, optional): Specifies how the images should be reshaped to the appropriate size.
        trainsform_train (str, optional): level of transformation for the training set.
        transform_validation (str, optional): level of transformation for the validation set.
        transform_test (str, optional): level of transformation for the test set.
        mean (float, optional): The mean value for normalizing pixel intensities.
        std (float, optional): The standard deviation value for normalizing pixel intensities.

    Raises:
        ValueError: If the ´resize´ is neither 'crop', 'resize' or 'padding'.
    
    Example:
        my_augmentation = Augmentation(resize = 'crop', transform_train = 'heavy')
    """
    
    def __init__(self, 
                 resize = 'crop', 
                 transform_train = 'moderate', 
                 transform_validation = None,
                 transform_test = None,
                 mean: float = 0.5, 
                 std: float = 0.5
                ):

        if resize not in ['crop', 'reshape', 'padding']:
            raise ValueError(f"Invalid resize type: {resize}. Use either 'crop', 'resize' or 'padding'.")

        self.resize = resize
        self.mean = mean
        self.std = std
        self.transform_train = transform_train
        self.transform_validation = transform_validation
        self.transform_test = transform_test
    
    def augment(self, im_h, im_w, level=None):
        """
        Returns an albumentations.core.composition.Compose class depending on the augmentation level.
        A baseline augmentation is implemented regardless of the level, and a set of augmentations are added depending of the level.
        The A.Resize() function is used if the user has specified a 'resize' int or tuple at the creation of the Augmentation class.

        Args:
            im_h (int): image height for resize.
            im_w (int): image width for resize.
            level (str, optional): level of augmentation.

        Raises:
            ValueError: If `level` is neither None, light, moderate nor heavy.
        """
        import albumentations as A
        from albumentations.pytorch import ToTensorV2

        # Check if one of standard augmentation levels
        if level not in [None,'light','moderate','heavy']:
            raise ValueError(f"Invalid transformation level: {level}. Please choose one of the following levels: None, 'light', 'moderate', 'heavy'.")

        # Baseline
        baseline_aug = [
            A.Normalize(mean = (self.mean),std = (self.std)),
            ToTensorV2()
        ]
        if self.resize == 'crop':
            resize_aug = [
                A.CenterCrop(im_h,im_w)
            ]
        elif self.resize == 'reshape':
            resize_aug =[
                A.Resize(im_h,im_w)
            ]
        elif self.resize == 'padding':
            resize_aug = [
                A.PadIfNeeded(im_h,im_w,border_mode = 0) # OpenCV border mode
            ]
        
        # Level of augmentation
        if level == None:
            level_aug = []
        elif level == 'light':
            level_aug = [
                A.RandomRotate90()
            ]
        elif level == 'moderate':
            level_aug = [
                A.RandomRotate90(),
                A.HorizontalFlip(p = 0.3),
                A.VerticalFlip(p = 0.3),
                A.GlassBlur(sigma = 0.7, p = 0.1),
                A.Affine(scale = [0.9,1.1], translate_percent = (0.1,0.1))
            ]
        elif level == 'heavy':
            level_aug = [
                A.RandomRotate90(),
                A.HorizontalFlip(p = 0.7),
                A.VerticalFlip(p = 0.7),
                A.GlassBlur(sigma = 1.2, iterations = 2, p = 0.3),
                A.Affine(scale = [0.8,1.4], translate_percent = (0.2,0.2), shear = (-15,15))
            ]

        augment = A.Compose(level_aug + resize_aug + baseline_aug)
        
        return augment


from .unet import UNet, Hyperparameters
from .augmentations import Augmentation
from .data import Dataset, prepare_dataloaders, prepare_datasets
from .models import inference, model_summary, train_model



from pathlib import Path
from abc import abstractmethod, ABC
from os import path

import gradio as gr

from .qim_theme import QimTheme
import qim3d.gui


# TODO: when offline it throws an error in cli
class BaseInterface(ABC):
    """
    Annotation tool and Data explorer as those don't need any examples.
    """

    def __init__(
        self,
        title: str,
        height: int,
        width: int = "100%",
        verbose: bool = False,
        custom_css: str = None,
    ):
        """
        title: Is displayed in tab
        height, width: If inline in launch method is True, sets the paramters of the widget. Inline defaults to True in py notebooks, otherwise is False
        verbose: If True, updates are printed into terminal
        custom_css: Only the name of the file in the css folder.
        """

        self.title = title
        self.height = height
        self.width = width
        self.verbose = bool(verbose)
        self.interface = None

        self.qim_dir = Path(qim3d.__file__).parents[0]
        self.custom_css = (
            path.join(self.qim_dir, "css", custom_css)
            if custom_css is not None
            else None
        )

    def set_visible(self):
        return gr.update(visible=True)

    def set_invisible(self):
        return gr.update(visible=False)

    def launch(self, img=None, force_light_mode: bool = True, **kwargs):
        """
        img: If None, user can upload image after the interface is launched.
            If defined, the interface will be launched with the image already there
            This argument is used especially in jupyter notebooks, where you can launch
            interface in loop with different picture every step
        force_light_mode: The qim platform doesn't have night mode. The qim_theme thus
            has option to display only light mode so it corresponds with the website. Preferably
            will be removed as we add night mode to the website.
        """

        # Create gradio interface
        if img is not None:
            self.img = img
        self.interface = self.create_interface(force_light_mode=force_light_mode)

        self.interface.launch(
            quiet=not self.verbose,
            height=self.height,
            width=self.width,
            favicon_path=Path(qim3d.__file__).parents[0]
            / "../docs/assets/qim3d-icon.svg",
            **kwargs,
        )

    def clear(self):
        """Used to reset outputs with the clear button"""
        return None

    def create_interface(self, force_light_mode: bool = True, **kwargs):
        # kwargs["img"] = self.img
        with gr.Blocks(
            theme=qim3d.gui.QimTheme(force_light_mode=force_light_mode),
            title=self.title,
            css=self.custom_css,
        ) as gradio_interface:
            gr.Markdown(f"# {self.title}")
            self.define_interface(**kwargs)
        return gradio_interface

    @abstractmethod
    def define_interface(self, **kwargs):
        pass

    def run_interface(self, host: str = "0.0.0.0"):
        qim3d.gui.run_gradio_app(self.create_interface(), host)


class InterfaceWithExamples(BaseInterface):
    """
    For Iso3D and Local Thickness
    """

    def __init__(
        self,
        title: str,
        height: int,
        width: int,
        verbose: bool = False,
        custom_css: str = None,
    ):
        super().__init__(title, height, width, verbose, custom_css)
        self._set_examples_list()

    def _set_examples_list(self):
        examples = [
            "blobs_256x256x256.tif",
            "fly_150x256x256.tif",
            "cement_128x128x128.tif",
            "NT_10x200x100.tif",
            "NT_128x128x128.tif",
            "shell_225x128x128.tif",
            "bone_128x128x128.tif",
        ]
        self.img_examples = []
        for example in examples:
            self.img_examples.append([path.join(self.qim_dir, "img_examples", example)])



from . import colormaps
from .cc import plot_cc
from .detection import circles
from .explore import (
    interactive_fade_mask,
    orthogonal,
    slicer,
    slices,
)
from .k3d import vol
from .local_thickness_ import local_thickness
from .structure_tensor import vectors
from .metrics import plot_metrics, grid_overview, grid_pred, vol_masked
from .preview import image_preview



import numpy as np
from scipy.ndimage import find_objects, label
from qim3d.utils.logger import log


class CC:
    def __init__(self, connected_components, num_connected_components):
        """
        Initializes a ConnectedComponents object.

        Args:
            connected_components (np.ndarray): The connected components.
            num_connected_components (int): The number of connected components.
        """
        self._connected_components = connected_components
        self.cc_count = num_connected_components
        
        self.shape = connected_components.shape
    
    def __len__(self):
        """
        Returns the number of connected components in the object.
        """
        return self.cc_count

    def get_cc(self, index=None, crop=False):
        """
        Get the connected component with the given index, if index is None selects a random component.

        Args:
            index (int): The index of the connected component. 
                            If none returns all components.
                            If 'random' returns a random component.
            crop (bool): If True, the volume is cropped to the bounding box of the connected component.

        Returns:
            np.ndarray: The connected component as a binary mask.
        """
        if index is None:
            volume = self._connected_components
        elif index == "random":
            index = np.random.randint(1, self.cc_count + 1)
            volume = self._connected_components == index
        else:
            assert 1 <= index <= self.cc_count, "Index out of range. Needs to be in range [1, cc_count]."
            volume = self._connected_components == index
            
        if crop:
            # As we index get_bounding_box element 0 will be the bounding box for the connected component at index
            bbox = self.get_bounding_box(index)[0] 
            volume = volume[bbox]
        
        return volume

    def get_bounding_box(self, index=None):
        """Get the bounding boxes of the connected components.

        Args:
            index (int, optional): The index of the connected component. If none selects all components.

        Returns:
            list: A list of bounding boxes.
        """

        if index:
            assert 1 <= index <= self.cc_count, "Index out of range."
            return find_objects(self._connected_components == index)
        else:
            return find_objects(self._connected_components)


def get_3d_cc(image: np.ndarray) -> CC:
    """ Returns an object (CC) containing the connected components of the input volume. Use plot_cc to visualize the connected components.

    Args:
        image (np.ndarray): An array-like object to be labeled. Any non-zero values in `input` are
            counted as features and zero values are considered the background.

    Returns:
        CC: A ConnectedComponents object containing the connected components and the number of connected components.

    Example:
        ```python
        import qim3d
        vol = qim3d.examples.cement_128x128x128[50:150]<60
        cc = qim3d.processing.get_3d_cc(vol)
        ```
    """
    connected_components, num_connected_components = label(image)
    log.info(f"Total number of connected components found: {num_connected_components}")
    return CC(connected_components, num_connected_components)



"""Wrapper for the local thickness function from the localthickness package including visualization functions."""

import numpy as np
from typing import Optional
from qim3d.utils.logger import log
import qim3d


def local_thickness(
    image: np.ndarray,
    scale: float = 1,
    mask: Optional[np.ndarray] = None,
    visualize=False,
    **viz_kwargs
) -> np.ndarray:
    """Wrapper for the local thickness function from the [local thickness package](https://github.com/vedranaa/local-thickness)

    The "Fast Local Thickness" by Vedrana Andersen Dahl and Anders Bjorholm Dahl from the Technical University of Denmark is a efficient algorithm for computing local thickness in 2D and 3D images.
    Their method significantly reduces computation time compared to traditional algorithms by utilizing iterative dilation with small structuring elements, rather than the large ones typically used.
    This approach allows the local thickness to be determined much faster, making it feasible for high-resolution volumetric data that are common in contemporary 3D microscopy.

    Testing against conventional methods and other Python-based tools like PoreSpy shows that the new algorithm is both accurate and faster, offering significant improvements in processing time for large datasets.


    Args:
        image (np.ndarray): 2D or 3D NumPy array representing the image/volume.
            If binary, it will be passed directly to the local thickness function.
            If grayscale, it will be binarized using Otsu's method.
        scale (float, optional): Downscaling factor, e.g. 0.5 for halving each dim of the image.
            Default is 1.
        mask (np.ndarray, optional): binary mask of the same size of the image defining parts of the
            image to be included in the computation of the local thickness. Default is None.
        visualize (bool, optional): Whether to visualize the local thickness. Default is False.
        **viz_kwargs: Additional keyword arguments passed to `qim3d.viz.local_thickness`. Only used if `visualize=True`.

    Returns:
        local_thickness (np.ndarray): 2D or 3D NumPy array representing the local thickness of the input image/volume.

    Example:
        ```python
        import qim3d

        fly = qim3d.examples.fly_150x256x256 # 3D volume
        lt_fly = qim3d.processing.local_thickness(fly, visualize=True, axis=0)
        ```
        ![local thickness 3d](assets/screenshots/local_thickness_3d.gif)

        ```python
        import qim3d

        blobs = qim3d.examples.blobs_256x256 # 2D image
        lt_blobs = qim3d.processing.local_thickness(blobs, visualize=True)
        ```
        ![local thickness 2d](assets/screenshots/local_thickness_2d.png)

    !!! info "Runtime and memory usage of the local thickness method for different volume sizes"
        ![local thickness estimate time and mem](assets/screenshots/Local_thickness_time_mem_estimation.png)

        Performance computed on Intel(R) Xeon(R) Gold 6226 CPU @ 2.70GHz.

    !!! quote "Reference"
        Dahl, V. A., & Dahl, A. B. (2023, June). Fast Local Thickness. 2023 IEEE/CVF Conference on Computer Vision and Pattern Recognition Workshops (CVPRW).
        <https://doi.org/10.1109/cvprw59228.2023.00456>

        ```bibtex
        @inproceedings{Dahl_2023, title={Fast Local Thickness},
        url={http://dx.doi.org/10.1109/CVPRW59228.2023.00456},
        DOI={10.1109/cvprw59228.2023.00456},
        booktitle={2023 IEEE/CVF Conference on Computer Vision and Pattern Recognition Workshops (CVPRW)},
        publisher={IEEE},
        author={Dahl, Vedrana Andersen and Dahl, Anders Bjorholm},
        year={2023},
        month=jun }

        ```



    """
    import localthickness as lt
    from skimage.filters import threshold_otsu

    # Check if input is binary
    if np.unique(image).size > 2:
        # If not, binarize it using Otsu's method, log the threshold and compute the local thickness
        threshold = threshold_otsu(image=image)
        log.warning(
            "Input image is not binary. It will be binarized using Otsu's method with threshold: {}".format(
                threshold
            )
        )
        local_thickness = lt.local_thickness(image > threshold, scale=scale, mask=mask)
    else:
        # If it is binary, compute the local thickness directly
        local_thickness = lt.local_thickness(image, scale=scale, mask=mask)

    # Visualize the local thickness if requested
    if visualize:
        display(qim3d.viz.local_thickness(image, local_thickness, **viz_kwargs))

    return local_thickness



"""qim3d: A Python package for 3D image processing and visualization.

The qim3d library is designed to make it easier to work with 3D imaging data in Python. 
It offers a range of features, including data loading and manipulation,
image processing and filtering, visualization of 3D data, and analysis of imaging results.

Documentation available at https://platform.qim.dk/qim3d/

"""

__version__ = "0.4.0"


import importlib as _importlib


class _LazyLoader:
    """Lazy loader to load submodules only when they are accessed"""

    def __init__(self, module_name):
        self.module_name = module_name
        self.module = None

    def _load(self):
        if self.module is None:
            self.module = _importlib.import_module(self.module_name)
        return self.module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)


# List of submodules
_submodules = [
    "examples",
    "generate",
    "gui",
    "io",
    "models",
    "processing",
    "tests",
    "utils",
    "viz",
    "cli",
]

# Creating lazy loaders for each submodule
for submodule in _submodules:
    globals()[submodule] = _LazyLoader(f"qim3d.{submodule}")



"""Wrapper for the structure tensor function from the structure_tensor package"""

from typing import Tuple
import logging
import numpy as np


def structure_tensor(
    vol: np.ndarray,
    sigma: float = 1.0,
    rho: float = 6.0,
    base_noise: bool = True,
    full: bool = False,
    visualize=False,
    **viz_kwargs
) -> Tuple[np.ndarray, np.ndarray]:
    """Wrapper for the 3D structure tensor implementation from the [structure_tensor package](https://github.com/Skielex/structure-tensor/).


    The structure tensor algorithm is a method for analyzing the orientation of fiber-like structures in 3D images.
    The core of the algorithm involves computing a 3-by-3 matrix at each point in a volume, capturing the local orientation. This matrix, known as the structure tensor, is derived from the gradients of the image intensities and integrates neighborhood information using Gaussian kernels.

    The implementation here used allows for fast and efficient computation using GPU-based processing, making it suitable for large datasets.
    This efficiency is particularly advantageous for high-resolution imaging techniques like X-ray computed microtomography (μCT).

    Args:
        vol (np.ndarray): 3D NumPy array representing the volume.
        sigma (float, optional): A noise scale, structures smaller than sigma will be removed by smoothing.
        rho (float, optional): An integration scale giving the size over the neighborhood in which the orientation is to be analysed.
        base_noise (bool, optional): A flag indicating whether to add a small noise to the volume. Default is True.
        full (bool, optional): A flag indicating that all three eigenvalues should be returned. Default is False.
        visualize (bool, optional): Whether to visualize the structure tensor. Default is False.
        **viz_kwargs: Additional keyword arguments for passed to `qim3d.viz.vectors`. Only used if `visualize=True`.

    Raises:
        ValueError: If the input volume is not 3D.

    Returns:
        val: An array with shape `(3, *vol.shape)` containing the eigenvalues of the structure tensor.
        vec: An array with shape `(3, *vol.shape)` if `full` is `False`, otherwise `(3, 3, *vol.shape)` containing eigenvectors.

    Example:
        ```python
        import qim3d

        vol = qim3d.examples.NT_128x128x128
        val, vec = qim3d.processing.structure_tensor(vol, visualize = True, axis = 2)
        ```
        ![structure tensor](assets/screenshots/structure_tensor_visualization.gif)


    !!! info "Runtime and memory usage of the structure tensor method for different volume sizes"
        ![structure tensor estimate time and mem](assets/screenshots/Structure_tensor_time_mem_estimation.png)

        Performance computed on Intel(R) Xeon(R) Gold 6226 CPU @ 2.70GHz.


    !!! quote "Reference"
        Jeppesen, N., et al. "Quantifying effects of manufacturing methods on fiber orientation in unidirectional composites using structure tensor analysis." Composites Part A: Applied Science and Manufacturing 149 (2021): 106541.
        <https://doi.org/10.1016/j.compositesa.2021.106541>

        ```bibtex
        @article{JEPPESEN2021106541,
        title = {Quantifying effects of manufacturing methods on fiber orientation in unidirectional composites using structure tensor analysis},
        journal = {Composites Part A: Applied Science and Manufacturing},
        volume = {149},
        pages = {106541},
        year = {2021},
        issn = {1359-835X},
        doi = {https://doi.org/10.1016/j.compositesa.2021.106541},
        url = {https://www.sciencedirect.com/science/article/pii/S1359835X21002633},
        author = {N. Jeppesen and L.P. Mikkelsen and A.B. Dahl and A.N. Christensen and V.A. Dahl}
        }

        ```

    """
    previous_logging_level = logging.getLogger().getEffectiveLevel()
    logging.getLogger().setLevel(logging.CRITICAL)
    import structure_tensor as st

    logging.getLogger().setLevel(previous_logging_level)

    if vol.ndim != 3:
        raise ValueError("The input volume must be 3D")

    # Ensure volume is a float
    if vol.dtype != np.float32 and vol.dtype != np.float64:
        vol = vol.astype(np.float32)

    if base_noise:
        # Add small noise to the volume
        # FIXME: This is a temporary solution to avoid uniform regions with constant values
        # in the volume, which lead to numerical issues in the structure tensor computation
        vol_noisy = vol + np.random.default_rng(seed=0).uniform(
            0, 1e-10, size=vol.shape
        )

        # Compute the structure tensor (of volume with noise)
        s_vol = st.structure_tensor_3d(vol_noisy, sigma, rho)

    else:
        # Compute the structure tensor (of volume without noise)
        s_vol = st.structure_tensor_3d(vol, sigma, rho)

    # Compute the eigenvalues and eigenvectors of the structure tensor
    val, vec = st.eig_special_3d(s_vol, full=full)

    if visualize:
        from qim3d.viz.structure_tensor import vectors

        display(vectors(vol, vec, **viz_kwargs))

    return val, vec



from fastapi import FastAPI
import qim3d.utils
from . import data_explorer
from . import iso3d
from . import local_thickness
from . import annotation_tool
from .qim_theme import QimTheme


def run_gradio_app(gradio_interface, host="0.0.0.0"):
    import gradio as gr
    import uvicorn

    # Get port using the QIM API
    port_dict = qim3d.utils.get_port_dict()

    if "gradio_port" in port_dict:
        port = port_dict["gradio_port"]
    elif "port" in port_dict:
        port = port_dict["port"]
    else:
        raise Exception("Port not specified from QIM API")

    qim3d.utils.gradio_header(gradio_interface.title, port)

    # Create FastAPI with mounted gradio interface
    app = FastAPI()
    path = f"/gui/{port_dict['username']}/{port}/"
    app = gr.mount_gradio_app(app, gradio_interface, path=path)

    # Full path
    print(f"http://{host}:{port}{path}")

    # Run the FastAPI server usign uvicorn
    uvicorn.run(app, host=host, port=int(port))



"Helper functions for testing"

import os
import matplotlib
import matplotlib.pyplot as plt
from pathlib import Path
import shutil
from PIL import Image
import socket
import numpy as np
from qim3d.utils.logger import log


def mock_plot():
    """Creates a mock plot of a sine wave.

    Returns:
        matplotlib.figure.Figure: The generated plot figure.

    Example:
        Creates a mock plot of a sine wave and displays the plot using `plt.show()`.

        >>> fig = mock_plot()
        >>> plt.show()
    """

    matplotlib.use("Agg")

    fig = plt.figure(figsize=(5, 4))
    axes = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    values = np.arange(0, 2 * np.pi, 0.01)
    axes.plot(values, np.sin(values))

    return fig


def mock_write_file(path, content="File created by qim3d"):
    """
    Creates a file at the specified path and writes a predefined text into it.

    Args:
        path (str): The path to the file to be created.

    Example:
        >>> mock_write_file("example.txt")
    """
    _file = open(path, "w", encoding="utf-8")
    _file.write(content)
    _file.close()


def is_server_running(ip, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        s.shutdown(2)
        return True
    except:
        return False


def temp_data(folder, remove=False, n=3, img_shape=(32, 32)):
    """Creates a temporary folder to test deep learning tools.

    Creates two folders, 'train' and 'test', who each also have two subfolders 'images' and 'labels'.
    n random images are then added to all four subfolders.
    If the 'remove' variable is True, the folders and their content are removed.

    Args:
        folder (str): The path where the folders should be placed.
        remove (bool, optional): If True, all folders are removed from their location.
        n (int, optional): Number of random images and labels in the temporary dataset.
        img_shape (tuple, options): Tuple with the height and width of the images and labels.

    Example:
        >>> tempdata('temporary_folder',n = 10, img_shape = (16,16))
    """
    folder_trte = ["train", "test"]
    sub_folders = ["images", "labels"]

    # Creating train/test folder
    path_train = Path(folder) / folder_trte[0]
    path_test = Path(folder) / folder_trte[1]

    # Creating folders for images and labels
    path_train_im = path_train / sub_folders[0]
    path_train_lab = path_train / sub_folders[1]
    path_test_im = path_test / sub_folders[0]
    path_test_lab = path_test / sub_folders[1]

    # Random image
    img = np.random.randint(2, size=img_shape, dtype=np.uint8)
    img = Image.fromarray(img)

    if not os.path.exists(path_train):
        os.makedirs(path_train_im)
        os.makedirs(path_test_im)
        os.makedirs(path_train_lab)
        os.makedirs(path_test_lab)
        for i in range(n):
            img.save(path_train_im / f"img_train{i}.png")
            img.save(path_train_lab / f"img_train{i}.png")
            img.save(path_test_im / f"img_test{i}.png")
            img.save(path_test_lab / f"img_test{i}.png")

    if remove:
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                log.warning("Failed to delete %s. Reason: %s" % (file_path, e))

        os.rmdir(folder)



""" Deals with DOI for references """
import json
import requests
from qim3d.utils.logger import log


def _validate_response(response):
    # Check if we got a good response
    if not response.ok:
        log.error(f"Could not read the provided DOI ({response.reason})")
        return False

    return True


def _doi_to_url(doi):
    if doi[:3] != "http":
        url = "https://doi.org/" + doi
    else:
        url = doi

    return url


def _make_request(doi, header):
    # Get url from doi
    url = _doi_to_url(doi)

    # run the request
    response = requests.get(url, headers=header, timeout=10)

    if not _validate_response(response):
        return None

    return response


def _log_and_get_text(doi, header):
    response = _make_request(doi, header)

    if response and response.encoding:
        # Explicitly decode the response content using the specified encoding
        text = response.content.decode(response.encoding)
        log.info(text)
        return text
    elif response:
        # If encoding is not specified, default to UTF-8
        text = response.text
        log.info(text)
        return text


def get_bibtex(doi):
    """Generates bibtex from doi"""
    header = {"Accept": "application/x-bibtex"}

    return _log_and_get_text(doi, header)

def cusom_header(doi, header):
    """Allows a custom header to be passed

    For example:
        doi = "https://doi.org/10.1101/2022.11.08.515664"
        header = {"Accept": "text/bibliography"}
        response = qim3d.utils.doi.cusom_header(doi, header)

    """
    return _log_and_get_text(doi, header)

def get_metadata(doi):
    """Generates a metadata dictionary from doi"""
    header = {"Accept": "application/vnd.citationstyles.csl+json"}
    response = _make_request(doi, header)

    metadata = json.loads(response.text)

    return metadata

def get_reference(doi):
    """Generates a metadata dictionary from doi and use it to build a reference string"""

    metadata = get_metadata(doi)
    reference_string = build_reference_string(metadata)

    return reference_string

def build_reference_string(metadata):
    """Generates a reference string from metadata"""
    authors = ", ".join([f"{author['family']} {author['given']}" for author in metadata['author']])
    year = metadata['issued']['date-parts'][0][0]
    title = metadata['title']
    publisher = metadata['publisher']
    url = metadata['URL']
    doi = metadata['DOI']

    reference_string = f"{authors} ({year}). {title}. {publisher} ({url}). DOI: {doi}"

    return reference_string



""" Dataset synchronization tasks """
import os
import subprocess
import outputformat as ouf
from qim3d.utils.logger import log
from pathlib import Path


class Sync:
    """Class for dataset synchronization tasks"""

    def __init__(self):
        # Checks if rsync is available
        if not self._check_rsync():
            raise RuntimeError(
                "Could not find rsync, please check if it is installed in your system."
            )

    def _check_rsync(self):
        """Check if rsync is available"""
        try:
            subprocess.run(["rsync", "--version"], capture_output=True, check=True)
            return True

        except Exception as error:
            log.error("rsync is not available")
            log.error(error)

            return False

    def check_destination(self, source, destination, checksum=False, verbose=True):
        """Check if all files from 'source' are in 'destination'

        This function compares the files in the 'source' directory to those in
        the 'destination' directory and reports any differences or missing files.

        Args:
            source (str or Path): The source directory path.
            destination (str or Path): The destination directory path.
            checksum (bool, optional): If True, use checksums to compare files (slower but more accurate).
                Default is False.
            verbose (bool, optional): If True, display a list of differing or missing files in the log.
                Default is True.

        Returns:
            list: A list of differing or missing file paths in the destination directory.

        """

        source = Path(source)
        destination = Path(destination)

        if checksum:
            rsync_args = "-avrc"
        else:
            rsync_args = "-avr"

        command = [
            "rsync",
            "-n",
            rsync_args,
            str(source) + os.path.sep,
            str(destination) + os.path.sep,
        ]

        out = subprocess.run(
            command,
            capture_output=True,
            check=True,
        )

        diff_files_and_folders = out.stdout.decode().splitlines()[1:-3]
        diff_files = [f for f in diff_files_and_folders if not f.endswith("/")]

        if len(diff_files) > 0 and verbose:
            title = "Source files differing or missing in destination"
            log.info(
                ouf.showlist(diff_files, style="line", return_str=True, title=title)
            )

        return diff_files

    def compare_dirs(self, source, destination, checksum=False, verbose=True):
        """Checks whether 'source' and 'destination' directories are synchronized.

        This function compares the contents of two directories
        ('source' and 'destination') and reports any differences.
        It checks for files that exist in one directory but not the other and
        files that are present in both but not equal.

        If no differences are found between the directories,
        it logs a message indicating that they are synchronized.
        If differences are found, it logs detailed information about the differing files.

        Args:
            source (str or Path): The source directory path.
            destination (str or Path): The destination directory path.
            checksum (bool, optional): If True, use checksums to compare files (slower but more accurate).
                Default is False.
            verbose (bool, optional): If True, display information about the comparison in the log.
                Default is True.

        Returns:
            None: This function does not return a value.

        """
        if verbose:
            s_files, s_dirs = self.count_files_and_dirs(source)
            d_files, d_dirs = self.count_files_and_dirs(destination)
            log.info("\n")

        s_d = self.check_destination(
            source, destination, checksum=checksum, verbose=False
        )
        d_s = self.check_destination(
            destination, source, checksum=checksum, verbose=False
        )

        if len(s_d) == 0 and len(d_s) == 0:
            # No differences
            if verbose:
                log.info(
                    "Source and destination are synchronized, no differences found."
                )
            return

        union = list(set(s_d + d_s))
        log.info(
            ouf.showlist(
                union,
                style="line",
                return_str=True,
                title=f"{len(union)} files are not in sync",
            )
        )

        intersection = list(set(s_d) & set(d_s))
        if len(intersection) > 0:
            log.info(
                ouf.showlist(
                    intersection,
                    style="line",
                    return_str=True,
                    title=f"{len(intersection)} files present on both, but not equal",
                )
            )

        s_exclusive = list(set(s_d).symmetric_difference(set(intersection)))
        if len(s_exclusive) > 0:
            log.info(
                ouf.showlist(
                    s_exclusive,
                    style="line",
                    return_str=True,
                    title=f"{len(s_exclusive)} files present only on {source}",
                )
            )

        d_exclusive = list(set(d_s).symmetric_difference(set(intersection)))
        if len(d_exclusive) > 0:
            log.info(
                ouf.showlist(
                    d_exclusive,
                    style="line",
                    return_str=True,
                    title=f"{len(d_exclusive)} files present only on {destination}",
                )
            )
        return

    def count_files_and_dirs(self, path, verbose=True):
        """Count the number of files and directories in the given path.

        This function recursively counts the number of files and
        directories in the specified directory 'path'.

        If 'verbose' is True, the function logs the total count
        of files and directories in the specified path.


        Args:
            path (str or Path): The directory path to count files and directories in.
            verbose (bool, optional): If True, display the total count in the log.
                Default is True.

        Returns:
            tuple: A tuple containing two values:
                - The count of files in the directory and its subdirectories.
                - The count of directories in the directory and its subdirectories.

        """
        path = Path(path)
        files = 0
        dirs = 0
        for p in os.scandir(path):
            if p.is_file():
                files += 1
            elif p.is_dir():
                dirs += 1
                file_count, dirs_count = self.count_files_and_dirs(p, verbose=False)
                files += file_count
                dirs += dirs_count

        if verbose:
            log.info(f"Total of {files} files and {dirs} directories on {path}")

        return files, dirs



from qim3d.utils.logger import log
import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, Union, Tuple
import ipywidgets as widgets

def local_thickness(
    image: np.ndarray,
    image_lt: np.ndarray,
    max_projection: bool = False,
    axis: int = 0,
    slice_idx: Optional[Union[int, float]] = None,
    show: bool = False,
    figsize: Tuple[int, int] = (15, 5),
) -> Union[plt.Figure, widgets.interactive]:
    """Visualizes the local thickness of a 2D or 3D image.

    Args:
        image (np.ndarray): 2D or 3D NumPy array representing the image/volume.
        image_lt (np.ndarray): 2D or 3D NumPy array representing the local thickness of the input
            image/volume.
        max_projection (bool, optional): If True, displays the maximum projection of the local
            thickness. Only used for 3D images. Defaults to False.
        axis (int, optional): The axis along which to visualize the local thickness.
            Unused for 2D images.
            Defaults to 0.
        slice_idx (int or float, optional): The initial slice to be visualized. The slice index
            can afterwards be changed. If value is an integer, it will be the index of the slice
            to be visualized. If value is a float between 0 and 1, it will be multiplied by the
            number of slices and rounded to the nearest integer. If None, the middle slice will
            be used for 3D images. Unused for 2D images. Defaults to None.
        show (bool, optional): If True, displays the plot (i.e. calls plt.show()). Defaults to False.
        figsize (Tuple[int, int], optional): The size of the figure. Defaults to (15, 5).

    Raises:
        ValueError: If the slice index is not an integer or a float between 0 and 1.

    Returns:
        If the input is 3D, returns an interactive widget. Otherwise, returns a matplotlib figure.

    Example:
        ```python
        import qim3d

        fly = qim3d.examples.fly_150x256x256 # 3D volume
        lt_fly = qim3d.processing.local_thickness(fly)
        qim3d.viz.local_thickness(fly, lt_fly, axis=0)
        ```
        ![local thickness 3d](assets/screenshots/local_thickness_3d.gif)

        
    """

    def _local_thickness(image, image_lt, show, figsize, axis=None, slice_idx=None):
        if slice_idx is not None:
            image = image.take(slice_idx, axis=axis)
            image_lt = image_lt.take(slice_idx, axis=axis)

        fig, axs = plt.subplots(1, 3, figsize=figsize, layout="constrained")

        axs[0].imshow(image, cmap="gray")
        axs[0].set_title("Original image")
        axs[0].axis("off")

        axs[1].imshow(image_lt, cmap="viridis")
        axs[1].set_title("Local thickness")
        axs[1].axis("off")

        plt.colorbar(
            axs[1].imshow(image_lt, cmap="viridis"), ax=axs[1], orientation="vertical"
        )

        axs[2].hist(image_lt[image_lt > 0].ravel(), bins=32, edgecolor="black")
        axs[2].set_title("Local thickness histogram")
        axs[2].set_xlabel("Local thickness")
        axs[2].set_ylabel("Count")

        if show:
            plt.show()

        plt.close()

        return fig

    # Get the middle slice if the input is 3D
    if len(image.shape) == 3:
        if max_projection:
            if slice_idx is not None:
                log.warning(
                    "slice_idx is not used for max_projection. It will be ignored."
                )
            image = image.max(axis=axis)
            image_lt = image_lt.max(axis=axis)
            return _local_thickness(image, image_lt, show, figsize)
        else:
            if slice_idx is None:
                slice_idx = image.shape[axis] // 2
            elif isinstance(slice_idx, float):
                if slice_idx < 0 or slice_idx > 1:
                    raise ValueError(
                        "Values of slice_idx of float type must be between 0 and 1."
                    )
                slice_idx = int(slice_idx * image.shape[0]) - 1
            slide_idx_slider = widgets.IntSlider(
                min=0,
                max=image.shape[axis] - 1,
                step=1,
                value=slice_idx,
                description="Slice index",
                layout=widgets.Layout(width="450px"),
            )
            widget_obj = widgets.interactive(
                _local_thickness,
                image=widgets.fixed(image),
                image_lt=widgets.fixed(image_lt),
                show=widgets.fixed(True),
                figsize=widgets.fixed(figsize),
                axis=widgets.fixed(axis),
                slice_idx=slide_idx_slider,
            )
            widget_obj.layout = widgets.Layout(align_items="center")
            if show:
                display(widget_obj)
            return widget_obj
    else:
        if max_projection:
            log.warning(
                "max_projection is only used for 3D images. It will be ignored."
            )
        if slice_idx is not None:
            log.warning("slice_idx is only used for 3D images. It will be ignored.")
        return _local_thickness(image, image_lt, show, figsize)


"""
This module provides a collection of colormaps useful for 3D visualization.
"""

import colorsys
from typing import Union, Tuple
import numpy as np
import math
from matplotlib.colors import LinearSegmentedColormap
from matplotlib import colormaps


def rearrange_colors(randRGBcolors_old, min_dist=0.5):
    # Create new list for re-arranged colors
    randRGBcolors_new = [randRGBcolors_old.pop(0)]

    while randRGBcolors_old:
        previous_color = randRGBcolors_new[-1]
        found = False

        # Find a next color that is at least min_dist away from previous color
        for color in randRGBcolors_old:
            if math.dist(color, previous_color) > min_dist:
                randRGBcolors_new.append(color)
                randRGBcolors_old.remove(color)
                found = True
                break

        # If no color was found, start over with the first color in the list
        if not found:
            randRGBcolors_new.append(randRGBcolors_old.pop(0))

    return randRGBcolors_new


def objects(
    nlabels: int,
    style: str = "bright",
    first_color_background: bool = True,
    last_color_background: bool = False,
    background_color: Union[Tuple[float, float, float], str] = (0.0, 0.0, 0.0),
    min_dist: int = 0.5,
    seed: int = 19,
) -> LinearSegmentedColormap:
    """
    Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks

    Args:
        nlabels (int): Number of labels (size of colormap).
        style (str, optional): 'bright' for strong colors, 'soft' for pastel colors, 'earth' for yellow/green/blue colors, 'ocean' for blue/purple/pink colors. Defaults to 'bright'.
        first_color_background (bool, optional): If True, the first color is used as background. Defaults to True.
        last_color_background (bool, optional): If True, the last color is used as background. Defaults to False.
        background_color (tuple or str, optional): RGB tuple or string for background color. Can be "black" or "white". Defaults to (0.0, 0.0, 0.0).
        min_dist (int, optional): Minimum distance between neighboring colors. Defaults to 0.5.
        seed (int, optional): Seed for random number generator. Defaults to 19.

    Returns:
        cmap (matplotlib.colors.LinearSegmentedColormap): Colormap for matplotlib


    Example:
        ```python
        import qim3d

        cmap_bright = qim3d.viz.colormaps.objects(nlabels=100, style = 'bright', first_color_background=True, background_color="black", min_dist=0.7)
        cmap_soft = qim3d.viz.colormaps.objects(nlabels=100, style = 'soft', first_color_background=True, background_color="black", min_dist=0.2)
        cmap_earth = qim3d.viz.colormaps.objects(nlabels=100, style = 'earth', first_color_background=True, background_color="black", min_dist=0.8)
        cmap_ocean = qim3d.viz.colormaps.objects(nlabels=100, style = 'ocean', first_color_background=True, background_color="black", min_dist=0.9)

        display(cmap_bright)
        display(cmap_soft)
        display(cmap_earth)
        display(cmap_ocean)
        ```
        ![colormap objects](assets/screenshots/viz-colormaps-objects-all.png)

        ```python
        import qim3d

        vol = qim3d.examples.cement_128x128x128
        binary = qim3d.processing.filters.gaussian(vol, 2) < 60
        labeled_volume, num_labels = qim3d.processing.operations.watershed(binary)

        cmap = qim3d.viz.colormaps.objects(num_labels, style = 'bright')
        qim3d.viz.slicer(labeled_volume, axis = 1, cmap=cmap)
        ```
        ![colormap objects](assets/screenshots/viz-colormaps-objects.gif)

    Tip:
        The `min_dist` parameter can be used to control the distance between neighboring colors.
        ![colormap objects mind_dist](assets/screenshots/viz-colormaps-min_dist.gif)


    """
    from skimage import color

    # Check style
    if style not in ("bright", "soft", "earth", "ocean"):
        raise ValueError(
            f'Please choose "bright", "soft", "earth" or "ocean" for style in qim3dCmap not "{style}"'
        )

    # Translate strings to background color
    color_dict = {"black": (0.0, 0.0, 0.0), "white": (1.0, 1.0, 1.0)}
    if not isinstance(background_color, tuple):
        try:
            background_color = color_dict[background_color]
        except KeyError:
            raise ValueError(
                f'Invalid color name "{background_color}". Please choose from {list(color_dict.keys())}.'
            )

    # Add one to nlabels to include the background color
    nlabels += 1

    # Create a new random generator, to locally set seed
    rng = np.random.default_rng(seed)

    # Generate color map for bright colors, based on hsv
    if style == "bright":
        randHSVcolors = [
            (
                rng.uniform(low=0.0, high=1),
                rng.uniform(low=0.4, high=1),
                rng.uniform(low=0.9, high=1),
            )
            for i in range(nlabels)
        ]

        # Convert HSV list to RGB
        randRGBcolors = []
        for HSVcolor in randHSVcolors:
            randRGBcolors.append(
                colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2])
            )

    # Generate soft pastel colors, by limiting the RGB spectrum
    if style == "soft":
        low = 0.6
        high = 0.95
        randRGBcolors = [
            (
                rng.uniform(low=low, high=high),
                rng.uniform(low=low, high=high),
                rng.uniform(low=low, high=high),
            )
            for i in range(nlabels)
        ]

    # Generate color map for earthy colors, based on LAB
    if style == "earth":
        randLABColors = [
            (
                rng.uniform(low=25, high=110),
                rng.uniform(low=-120, high=70),
                rng.uniform(low=-70, high=70),
            )
            for i in range(nlabels)
        ]

        # Convert LAB list to RGB
        randRGBcolors = []
        for LabColor in randLABColors:
            randRGBcolors.append(color.lab2rgb([[LabColor]])[0][0].tolist())

    # Generate color map for ocean colors, based on LAB
    if style == "ocean":
        randLABColors = [
            (
                rng.uniform(low=0, high=110),
                rng.uniform(low=-128, high=160),
                rng.uniform(low=-128, high=0),
            )
            for i in range(nlabels)
        ]

        # Convert LAB list to RGB
        randRGBcolors = []
        for LabColor in randLABColors:
            randRGBcolors.append(color.lab2rgb([[LabColor]])[0][0].tolist())

    # Re-arrange colors to have a minimum distance between neighboring colors
    randRGBcolors = rearrange_colors(randRGBcolors, min_dist)

    # Set first and last color to background
    if first_color_background:
        randRGBcolors[0] = background_color

    if last_color_background:
        randRGBcolors[-1] = background_color

    # Create colormap
    objects = LinearSegmentedColormap.from_list("objects", randRGBcolors, N=nlabels)

    return objects


qim = LinearSegmentedColormap.from_list(
    "qim",
    [
        (0.6, 0.0, 0.0),  # 990000
        (1.0, 0.6, 0.0),  # ff9900
    ],
)
"""
Defines colormap in QIM logo colors. Can be accessed as module attribute or easily by ```cmap = 'qim'```

Example:
    ```python

    import qim3d

    display(qim3d.viz.colormaps.qim)
    ```
    ![colormap objects](assets/screenshots/viz-colormaps-qim.png)
"""
colormaps.register(qim)



import numpy as np
import qim3d.processing.filters as filters
from qim3d.utils.logger import log


def remove_background(
    vol: np.ndarray,
    median_filter_size: int = 2,
    min_object_radius: int = 3,
    background: str = "dark",
    **median_kwargs
) -> np.ndarray:
    """
    Remove background from a volume using a qim3d filters.

    Args:
        vol (np.ndarray): The volume to remove background from.
        median_filter_size (int, optional): The size of the median filter. Defaults to 2.
        min_object_radius (int, optional): The radius of the structuring element for the tophat filter. Defaults to 3.
        background (str, optional): The background type. Can be 'dark' or 'bright'. Defaults to 'dark'.
        **median_kwargs: Additional keyword arguments for the Median filter.

    Returns:
        np.ndarray: The volume with background removed.


    Example:
        ```python
        import qim3d

        vol = qim3d.examples.cement_128x128x128
        qim3d.viz.slices(vol, vmin=0, vmax=255)
        ```
        ![operations-remove_background_before](assets/screenshots/operations-remove_background_before.png)  

        ```python
        vol_filtered  = qim3d.processing.operations.remove_background(vol,
                                                              min_object_radius=3, 
                                                              background="bright")
        qim3d.viz.slices(vol_filtered, vmin=0, vmax=255)
        ```
        ![operations-remove_background_after](assets/screenshots/operations-remove_background_after.png)
    """

    # Create a pipeline with a median filter and a tophat filter
    pipeline = filters.Pipeline(
        filters.Median(size=median_filter_size, **median_kwargs),
        filters.Tophat(radius=min_object_radius, background=background),
    )

    # Apply the pipeline to the volume
    return pipeline(vol)

def watershed(
        bin_vol: np.ndarray
) -> tuple[np.ndarray, int]:
    """
    Apply watershed segmentation to a binary volume.

    Args:
        bin_vol (np.ndarray): Binary volume to segment.
    
    Returns:
        tuple[np.ndarray, int]: Labeled volume after segmentation, number of objects found.

    Example:
        ```python
        import qim3d

        vol = qim3d.examples.cement_128x128x128
        binary = qim3d.processing.filters.gaussian(vol, 2)<60

        qim3d.viz.slices(binary, axis=1)
        ```
        ![operations-watershed_before](assets/screenshots/operations-watershed_before.png)  

        ```python
        labeled_volume, num_labels = qim3d.processing.operations.watershed(binary)

        cmap = qim3d.viz.colormaps.objects(num_labels)
        qim3d.viz.slices(labeled_volume, axis = 1, cmap=cmap)
        ```
        ![operations-watershed_after](assets/screenshots/operations-watershed_after.png)  

    """
    import skimage
    import scipy

    # Compute distance transform of binary volume
    distance= scipy.ndimage.distance_transform_edt(bin_vol)

    # Find peak coordinates in distance transform
    coords = skimage.feature.peak_local_max(distance, labels=bin_vol)

    # Create a mask with peak coordinates
    mask = np.zeros(distance.shape, dtype=bool)
    mask[tuple(coords.T)] = True

    # Label peaks
    markers, _ = scipy.ndimage.label(mask)

    # Apply watershed segmentation
    labeled_volume = skimage.segmentation.watershed(-distance, markers=markers, mask=bin_vol)
    
    # Extract number of objects found
    num_labels = len(np.unique(labeled_volume))-1
    log.info(f"Total number of objects found: {num_labels}")

    return labeled_volume, num_labels

def fade_mask(
    vol: np.ndarray,
    decay_rate: float = 10,
    ratio: float = 0.5,
    geometry: str = "spherical",
    invert: bool = False,
    axis: int = 0,
    ) -> np.ndarray:
    """
    Apply edge fading to a volume.

    Args:
        vol (np.ndarray): The volume to apply edge fading to.
        decay_rate (float, optional): The decay rate of the fading. Defaults to 10.
        ratio (float, optional): The ratio of the volume to fade. Defaults to 0.5.
        geometric (str, optional): The geometric shape of the fading. Can be 'spherical' or 'cylindrical'. Defaults to 'spherical'.
        axis (int, optional): The axis along which to apply the fading. Defaults to 0.
    
    Returns:
        vol_faded (np.ndarray): The volume with edge fading applied.

    Example:
        ```python
        import qim3d
        qim3d.viz.vol(vol)
        ```
        Image before edge fading has visible artifacts from the support. Which obscures the object of interest.
        ![operations-edge_fade_before](assets/screenshots/operations-edge_fade_before.png)  

        ```python
        import qim3d
        vol_faded = qim3d.processing.operations.edge_fade(vol, decay_rate=4, ratio=0.45, geometric='cylindrical')
        qim3d.viz.vol(vol_faded)
        ```
        Afterwards the artifacts are faded out, making the object of interest more visible for visualization purposes.
        ![operations-edge_fade_after](assets/screenshots/operations-edge_fade_after.png)

    """
    if 0 > axis or axis >= vol.ndim:
        raise ValueError("Axis must be between 0 and the number of dimensions of the volume")

    # Generate the coordinates of each point in the array
    shape = vol.shape
    z, y, x = np.indices(shape)
    
    # Calculate the center of the array
    center = np.array([(s - 1) / 2 for s in shape])
    
    # Calculate the distance of each point from the center
    if geometry == "spherical":
        distance = np.linalg.norm([z - center[0], y - center[1], x - center[2]], axis=0)
    elif geometry == "cylindrical":
        distance_list = np.array([z - center[0], y - center[1], x - center[2]])
        # remove the axis along which the fading is not applied
        distance_list = np.delete(distance_list, axis, axis=0)
        distance = np.linalg.norm(distance_list, axis=0)
    else:
        raise ValueError("geometric must be 'spherical' or 'cylindrical'")
    
    # Normalize the distances so that they go from 0 at the center to 1 at the farthest point
    max_distance = np.linalg.norm(center)
    normalized_distance = distance / (max_distance*ratio)
    
    # Apply the decay rate
    faded_distance = normalized_distance ** decay_rate
    
    # Invert the distances to have 1 at the center and 0 at the edges
    fade_array = 1 - faded_distance
    fade_array[fade_array<=0]=0
    
    if invert:
        fade_array = -(fade_array-1)

    # Apply the fading to the volume
    vol_faded = vol * fade_array

    return vol_faded

def overlay_rgb_images(
    background: np.ndarray, 
    foreground: np.ndarray, 
    alpha: float = 0.5
    ) -> np.ndarray:
    """
    Overlay an RGB foreground onto an RGB background using alpha blending.

    Args:
        background (numpy.ndarray): The background RGB image.
        foreground (numpy.ndarray): The foreground RGB image (usually masks).
        alpha (float, optional): The alpha value for blending. Defaults to 0.5.

    Returns:
        composite (numpy.ndarray): The composite RGB image with overlaid foreground.

    Raises:
        ValueError: If input images have different shapes.

    Note:
        - The function performs alpha blending to overlay the foreground onto the background.
        - It ensures that the background and foreground have the same shape before blending.
        - It calculates the maximum projection of the foreground and blends them onto the background.
        - Brightness outside the foreground is adjusted to maintain consistency with the background.
    """

    # Igonore alpha in case its there
    background = background[..., :3]
    foreground = foreground[..., :3]

    # Ensure both images have the same shape
    if background.shape != foreground.shape:
        raise ValueError("Input images must have the same shape")

    # Perform alpha blending
    foreground_max_projection = np.amax(foreground, axis=2)
    foreground_max_projection = np.stack((foreground_max_projection,) * 3, axis=-1)

    # Normalize if we have something
    if np.max(foreground_max_projection) > 0:
        foreground_max_projection = foreground_max_projection / np.max(
            foreground_max_projection
        )

    composite = background * (1 - alpha) + foreground * alpha
    composite = np.clip(composite, 0, 255).astype("uint8")

    # Adjust brightness outside foreground
    composite = composite + (background * (1 - alpha)) * (1 - foreground_max_projection)

    return composite.astype("uint8")


import matplotlib.pyplot as plt
import numpy as np

from qim3d.utils.logger import log
import qim3d.viz.colormaps


def plot_cc(
    connected_components,
    component_indexs: list | tuple = None,
    max_cc_to_plot=32,
    overlay=None,
    crop=False,
    show=True,
    **kwargs,
) -> list[plt.Figure]:
    """
    Plots the connected components from a `qim3d.processing.cc.CC` object. If an overlay image is provided, the connected component will be masked to the overlay image.

    Parameters:
        connected_components (CC): The connected components object.
        component_indexs (list | tuple, optional): The components to plot. If None the first max_cc_to_plot=32 components will be plotted. Defaults to None.
        max_cc_to_plot (int, optional): The maximum number of connected components to plot. Defaults to 32.
        overlay (optional): Overlay image. Defaults to None.
        crop (bool, optional): Whether to crop the image to the cc. Defaults to False.
        show (bool, optional): Whether to show the figure. Defaults to True.
        **kwargs: Additional keyword arguments to pass to `qim3d.viz.slices`.

    Returns:
        figs (list[plt.Figure]): List of figures, if `show=False`.

    Example:
        ```python
        import qim3d
        vol = qim3d.examples.cement_128x128x128[50:150]
        vol_bin = vol<80
        cc = qim3d.processing.get_3d_cc(vol_bin)
        qim3d.viz.plot_cc(cc, crop=True, show=True, overlay=None, n_slices=5, component_indexs=[4,6,7])
        qim3d.viz.plot_cc(cc, crop=True, show=True, overlay=vol, n_slices=5, component_indexs=[4,6,7])
        ```
        ![plot_cc_no_overlay](assets/screenshots/plot_cc_no_overlay.png)
        ![plot_cc_overlay](assets/screenshots/plot_cc_overlay.png)
    """
    # if no components are given, plot the first max_cc_to_plot=32 components
    if component_indexs is None:
        if len(connected_components) > max_cc_to_plot:
            log.warning(
                f"More than {max_cc_to_plot} connected components found. Only the first {max_cc_to_plot} will be plotted. Change max_cc_to_plot to plot more components."
            )
        component_indexs = range(
            1, min(max_cc_to_plot + 1, len(connected_components) + 1)
        )

    figs = []
    for component in component_indexs:
        if overlay is not None:
            assert (
                overlay.shape == connected_components.shape
            ), f"Overlay image must have the same shape as the connected components. overlay.shape=={overlay.shape} != connected_components.shape={connected_components.shape}."

            # plots overlay masked to connected component
            if crop:
                # Crop the overlay image based on the bounding box of the component
                bb = connected_components.get_bounding_box(component)[0]
                cc = connected_components.get_cc(component, crop=True)
                overlay_crop = overlay[bb]
                # use cc as mask for overlay_crop, where all values in cc set to 0 should be masked out, cc contains integers
                overlay_crop = np.where(cc == 0, 0, overlay_crop)
                fig = qim3d.viz.slices(overlay_crop, show=show, **kwargs)
            else:
                cc = connected_components.get_cc(component, crop=False)
                overlay_crop = np.where(cc == 0, 0, overlay)
                fig = qim3d.viz.slices(overlay_crop, show=show, **kwargs)
        else:
            # assigns discrete color map to each connected component if not given
            if "cmap" not in kwargs:
                kwargs["cmap"] = qim3d.viz.colormaps.objects(len(component_indexs))

            # Plot the connected component without overlay
            fig = qim3d.viz.slices(
                connected_components.get_cc(component, crop=crop), show=show, **kwargs
            )

        figs.append(fig)

    if not show:
        return figs

    return



"""Provides a custom Dataset class for building a PyTorch dataset."""
from pathlib import Path
from PIL import Image
from qim3d.utils.logger import log
import torch
import numpy as np


class Dataset(torch.utils.data.Dataset):
    """
    Custom Dataset class for building a PyTorch dataset.

    Args:
        root_path (str): The root directory path of the dataset.
        split (str, optional): The split of the dataset, either "train" or "test".
            Default is "train".
        transform (callable, optional): A callable function or transformation to 
            be applied to the data. Default is None.

    Raises:
        ValueError: If the provided split is not valid (neither "train" nor "test").

    Attributes:
        split (str): The split of the dataset ("train" or "test").
        transform (callable): The transformation to be applied to the data.
        sample_images (list): A list containing the paths to the sample images in the dataset.
        sample_targets (list): A list containing the paths to the corresponding target images
            in the dataset.

    Methods:
        __len__(): Returns the total number of samples in the dataset.
        __getitem__(idx): Returns the image and its target segmentation at the given index.

    Usage:
        dataset = Dataset(root_path="path/to/dataset", split="train", 
            transform=albumentations.Compose([ToTensorV2()]))
        image, target = dataset[idx]
    """
    def __init__(self, root_path: str, split="train", transform=None):
        super().__init__()

        # Check if split is valid
        if split not in ["train", "test"]:
            raise ValueError("Split must be either train or test")

        self.split = split
        self.transform = transform

        path = Path(root_path) / split

        self.sample_images = [file for file in sorted((path / "images").iterdir())]
        self.sample_targets = [file for file in sorted((path / "labels").iterdir())]
        assert len(self.sample_images) == len(self.sample_targets)

        # checking the characteristics of the dataset
        self.check_shape_consistency(self.sample_images)

    def __len__(self):
        return len(self.sample_images)

    def __getitem__(self, idx):
        image_path = self.sample_images[idx]
        target_path = self.sample_targets[idx]

        image = Image.open(str(image_path))
        image = np.array(image)
        target = Image.open(str(target_path))
        target = np.array(target)

        if self.transform:
            transformed = self.transform(image=image, mask=target)
            image = transformed["image"]
            target = transformed["mask"]

        return image, target


    # TODO: working with images of different sizes
    def check_shape_consistency(self,sample_images):
        image_shapes= []
        for image_path in sample_images:
            image_shape = self._get_shape(image_path)
            image_shapes.append(image_shape)

        # check if all images have the same size.
        consistency_check = all(i == image_shapes[0] for i in image_shapes)

        if not consistency_check:
            raise NotImplementedError(
                "Only images of all the same size can be processed at the moment"
            )
        else:
            log.debug(
                "Images are all the same size!"
            )
        return consistency_check
    
    def _get_shape(self,image_path):
        return Image.open(str(image_path)).size


def check_resize(im_height: int, im_width: int, resize: str, n_channels: int):
    """
    Checks the compatibility of the image shape with the depth of the model.
    If the image height and width cannot be divided by 2 `n_channels` times, then the image size is inappropriate.
    If so, the image is changed to the closest appropriate dimension, and the user is notified with a warning.

    Args:
        im_height (int) : Height of the original image from the dataset.
        im_width (int)  : Width of the original image from the dataset.
        resize (str)    : Type of resize to be used on the image if the size doesn't fit the model.
        n_channels (int): Number of channels in the model.

    Raises:
        ValueError: If the image size is smaller than minimum required for the model's depth.               
    """
    # finding suitable size to upsize with padding 
    if resize == 'padding':
        h_adjust, w_adjust = (im_height // 2**n_channels+1) * 2**n_channels , (im_width // 2**n_channels+1) * 2**n_channels
        
    # finding suitable size to downsize with crop / resize
    else:
        h_adjust, w_adjust = (im_height // 2**n_channels) * 2**n_channels , (im_width // 2**n_channels) * 2**n_channels    
    
    if h_adjust == 0 or w_adjust == 0:
        raise ValueError("The size of the image is too small compared to the depth of the UNet. Choose a different 'resize' and/or a smaller model.")
    
    elif h_adjust != im_height or w_adjust != im_width:
        log.warning(f"The image size doesn't match the Unet model's depth. The image is changed with '{resize}', from {im_height, im_width} to {h_adjust, w_adjust}.")
    
    return h_adjust, w_adjust 


def prepare_datasets(path: str, val_fraction: float, model, augmentation):
    """
    Splits and augments the train/validation/test datasets.

    Args:
        path (str): Path to the dataset.
        val_fraction (float): Fraction of the data for the validation set.
        model (torch.nn.Module): PyTorch Model.
        augmentation (albumentations.core.composition.Compose): Augmentation class for the dataset with predefined augmentation levels.

    Raises:
        ValueError: if the validation fraction is not a float, and is not between 0 and 1. 
    """
    
    if not isinstance(val_fraction,float) or not (0 <= val_fraction < 1):
        raise ValueError("The validation fraction must be a float between 0 and 1.")

    resize = augmentation.resize
    n_channels = len(model.channels)

    # taking the size of the 1st image in the dataset
    im_path = Path(path) / 'train'
    first_img = sorted((im_path / "images").iterdir())[0]
    image = Image.open(str(first_img))
    orig_h, orig_w = image.size[:2]
        
    final_h, final_w = check_resize(orig_h, orig_w, resize, n_channels)

    train_set = Dataset(root_path = path, transform = augmentation.augment(final_h, final_w, augmentation.transform_train))
    val_set   = Dataset(root_path = path, transform = augmentation.augment(final_h, final_w, augmentation.transform_validation))
    test_set  = Dataset(root_path = path, split='test', transform = augmentation.augment(final_h, final_w, augmentation.transform_test))

    split_idx = int(np.floor(val_fraction * len(train_set)))
    indices = torch.randperm(len(train_set))

    train_set = torch.utils.data.Subset(train_set, indices[split_idx:])
    val_set = torch.utils.data.Subset(val_set, indices[:split_idx])
    
    return train_set, val_set, test_set


def prepare_dataloaders(train_set, val_set, test_set, batch_size, shuffle_train = True, num_workers = 8, pin_memory = False):  
    """
    Prepares the dataloaders for model training.

    Args:
        train_set (torch.utils.data): Training dataset. 
        val_set (torch.utils.data):   Validation dataset.
        test_set (torch.utils.data):  Testing dataset.
        batch_size (int): Size of the batches that should be trained upon.
        shuffle_train (bool, optional): Optional input to shuffle the training data (training robustness).
        num_workers (int, optional): Defines how many processes should be run in parallel.
        pin_memory (bool, optional): Loads the datasets as CUDA tensors.
    """
    from torch.utils.data import DataLoader

    train_loader = DataLoader(dataset=train_set, batch_size=batch_size, shuffle=shuffle_train, num_workers=num_workers, pin_memory=pin_memory)
    val_loader = DataLoader(dataset=val_set, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
    test_loader = DataLoader(dataset=test_set, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)

    return train_loader,val_loader,test_loader


"""UNet model and Hyperparameters class."""

import torch.nn as nn

from qim3d.utils.logger import log


class UNet(nn.Module):
    """
    2D UNet model for QIM imaging.

    This class represents a 2D UNet model designed for imaging segmentation tasks.

    Args:
        size (str, optional): Size of the UNet model. Must be one of 'small', 'medium', or 'large'. Defaults to 'medium'.
        dropout (float, optional): Dropout rate between 0 and 1. Defaults to 0.
        kernel_size (int, optional): Convolution kernel size. Defaults to 3.
        up_kernel_size (int, optional): Up-convolution kernel size. Defaults to 3.
        activation (str, optional): Activation function. Defaults to 'PReLU'.
        bias (bool, optional): Whether to include bias in convolutions. Defaults to True.
        adn_order (str, optional): ADN (Activation, Dropout, Normalization) ordering. Defaults to 'NDA'.

    Raises:
        ValueError: If `size` is not one of 'small', 'medium', or 'large'.

    Example:
        ```python
        model = UNet(size='large')
        ```
    """

    def __init__(
        self,
        size="medium",
        dropout=0,
        kernel_size=3,
        up_kernel_size=3,
        activation="PReLU",
        bias=True,
        adn_order="NDA",
    ):
        super().__init__()
        if size not in ["small", "medium", "large"]:
            raise ValueError(
                f"Invalid model size: {size}. Size must be one of the following: 'small', 'medium', 'large'."
            )

        self.size = size
        self.dropout = dropout
        self.kernel_size = kernel_size
        self.up_kernel_size = up_kernel_size
        self.activation = activation
        self.bias = bias
        self.adn_order = adn_order

        self.model = self._model_choice()

    def _model_choice(self):
        from monai.networks.nets import UNet as monai_UNet

        if self.size == "small":
            self.channels = (64, 128, 256)
        elif self.size == "medium":
            self.channels = (64, 128, 256, 512, 1024)
        elif self.size == "large":
            self.channels = (64, 128, 256, 512, 1024, 2048)

        model = monai_UNet(
            spatial_dims=2,
            in_channels=1,  # TODO: check if image has 1 or multiple input channels
            out_channels=1,
            channels=self.channels,
            strides=(2,) * (len(self.channels) - 1),
            kernel_size=self.kernel_size,
            up_kernel_size=self.up_kernel_size,
            act=self.activation,
            dropout=self.dropout,
            bias=self.bias,
            adn_ordering=self.adn_order,
        )
        return model

    def forward(self, x):
        x = self.model(x)
        return x


class Hyperparameters:
    """
    Hyperparameters for QIM segmentation.

    Args:
        model (torch.nn.Module): PyTorch model.
        n_epochs (int, optional): Number of training epochs. Defaults to 10.
        learning_rate (float, optional): Learning rate for the optimizer. Defaults to 1e-3.
        optimizer (str, optional): Optimizer algorithm. Must be one of 'Adam', 'SGD', 'RMSprop'. Defaults to 'Adam'.
        momentum (float, optional): Momentum value for SGD and RMSprop optimizers. Defaults to 0.
        weight_decay (float, optional): Weight decay (L2 penalty) for the optimizer. Defaults to 0.
        loss_function (str, optional): Loss function criterion. Must be one of 'BCE', 'Dice', 'Focal', 'DiceCE'. Defaults to 'BCE'.

    Raises:
        ValueError: If `loss_function` is not one of 'BCE', 'Dice', 'Focal', 'DiceCE'.
        ValueError: If `optimizer` is not one of 'Adam', 'SGD', 'RMSprop'.

    Example:
        ```
        hyperparams = Hyperparameters(model=my_model, n_epochs=20, learning_rate=0.001)

        params_dict = hyperparams() # Get the hyperparameters

        optimizer = params_dict['optimizer']
        criterion = params_dict['criterion']
        n_epochs  = params_dict['n_epochs']
        ```
    """

    def __init__(
        self,
        model,
        n_epochs=10,
        learning_rate=1e-3,
        optimizer="Adam",
        momentum=0,
        weight_decay=0,
        loss_function="Focal",
    ):

        # TODO: implement custom loss_functions? then add a check to see if loss works for segmentation.
        if loss_function not in ["BCE", "Dice", "Focal", "DiceCE"]:
            raise ValueError(
                f"Invalid loss function: {loss_function}. Loss criterion must "
                "be one of the following: 'BCE','Dice','Focal','DiceCE'."
            )
        # TODO: implement custom optimizer? and add check to see if valid.
        if optimizer not in ["Adam", "SGD", "RMSprop"]:
            raise ValueError(
                f"Invalid optimizer: {optimizer}. Optimizer must "
                "be one of the following: 'Adam', 'SGD', 'RMSprop'."
            )

        if (momentum != 0) and optimizer == "Adam":
            log.info(
                "Momentum isn't an input in the 'Adam' optimizer. "
                "Change optimizer to 'SGD' or 'RMSprop' to use momentum."
            )

        self.model = model
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.momentum = momentum
        self.weight_decay = weight_decay
        self.loss_function = loss_function

    def __call__(self):
        return self.model_params(
            self.model,
            self.n_epochs,
            self.optimizer,
            self.learning_rate,
            self.weight_decay,
            self.momentum,
            self.loss_function,
        )

    def model_params(
        self,
        model,
        n_epochs,
        optimizer,
        learning_rate,
        weight_decay,
        momentum,
        loss_function,
    ):

        optim = self._optimizer(model, optimizer, learning_rate, weight_decay, momentum)
        criterion = self._loss_functions(loss_function)

        hyper_dict = {
            "optimizer": optim,
            "criterion": criterion,
            "n_epochs": n_epochs,
        }
        return hyper_dict

    # selecting the optimizer
    def _optimizer(self, model, optimizer, learning_rate, weight_decay, momentum):
        from torch.optim import Adam, SGD, RMSprop

        if optimizer == "Adam":
            optim = Adam(
                model.parameters(), lr=learning_rate, weight_decay=weight_decay
            )
        elif optimizer == "SGD":
            optim = SGD(
                model.parameters(),
                lr=learning_rate,
                momentum=momentum,
                weight_decay=weight_decay,
            )
        elif optimizer == "RMSprop":
            optim = RMSprop(
                model.parameters(),
                lr=learning_rate,
                weight_decay=weight_decay,
                momentum=momentum,
            )
        return optim

    # selecting the loss function
    def _loss_functions(self, loss_function):
        from monai.losses import FocalLoss, DiceLoss, DiceCELoss
        from torch.nn import BCEWithLogitsLoss

        if loss_function == "BCE":
            criterion = BCEWithLogitsLoss(reduction="mean")
        elif loss_function == "Dice":
            criterion = DiceLoss(sigmoid=True, reduction="mean")
        elif loss_function == "Focal":
            criterion = FocalLoss(reduction="mean")
        elif loss_function == "DiceCE":
            criterion = DiceCELoss(sigmoid=True, reduction="mean")
        return criterion



"Manages downloads and access to data"

import os
import urllib.request

from urllib.parse import quote
from tqdm import tqdm
from pathlib import Path

from qim3d.io import load
from qim3d.utils.logger import log
import outputformat as ouf


class Downloader:
    """Class for downloading large data files available on the [QIM data repository](https://data.qim.dk/).

    Attributes:
        [folder_name] (str): folder class with the name of the folder in <https://data.qim.dk/>

    Example:
        ```python
        import qim3d
        
        dl = qim3d.io.Downloader()
        img = dl.Corals.Coral2_DOWNSAMPLED(load_file = True)
        ```
    """

    def __init__(self):
        folders = _extract_names()
        for idx, folder in enumerate(folders):
            exec(f"self.{folder} = self._Myfolder(folder)")

    class _Myfolder:
        """Class for extracting the files from each folder in the Downloader class.

        Args:
            folder(str): name of the folder of interest in the QIM data repository.

        Methods:
             _make_fn(folder,file): creates custom functions for each file found in the folder.
            [file_name_1](load_file,optional): Function to download file number 1 in the given folder.
            [file_name_2](load_file,optional): Function to download file number 2 in the given folder.
            ...
            [file_name_n](load_file,optional): Function to download file number n in the given folder.
        """

        def __init__(self, folder):
            files = _extract_names(folder)

            for idx, file in enumerate(files):
                # Changes names to usable function name.
                file_name = file
                if ("%20" in file) or ("-" in file):
                    file_name = file_name.replace("%20", "_")
                    file_name = file_name.replace("-", "_")

                setattr(self, f'{file_name.split(".")[0]}', self._make_fn(folder, file))

        def _make_fn(self, folder, file):
            """Private method that returns a function. The function downloads the chosen file from the folder.

            Args:
                folder(str): Folder where the file is located.
                file(str): Name of the file to be downloaded.

            Returns:
                    function: the function used to download the file.
            """

            url_dl = "https://archive.compute.dtu.dk/download/public/projects/viscomp_data_repository"

            def _download(load_file=False, virtual_stack=True):
                """Downloads the file and optionally also loads it.

                Args:
                    load_file(bool,optional): Whether to simply download or also load the file.

                Returns:
                    virtual_stack: The loaded image.
                """

                download_file(url_dl, folder, file)
                if load_file == True:
                    log.info(f"\nLoading {file}")
                    file_path = os.path.join(folder, file)

                    return load(path=file_path, virtual_stack=virtual_stack)

            return _download


def _update_progress(pbar, blocknum, bs):
    """
    Helper function for the ´download_file()´ function. Updates the progress bar.
    """

    pbar.update(blocknum * bs - pbar.n)


def _get_file_size(url):
    """
    Helper function for the ´download_file()´ function. Finds the size of the file.
    """

    return int(urllib.request.urlopen(url).info().get("Content-Length", -1))


def download_file(path, name, file):
    """Downloads the file from path / name / file.

    Args:
        path(str): path to the folders available.
        name(str): name of the folder of interest.
        file(str): name of the file to be downloaded.
    """

    if not os.path.exists(name):
        os.makedirs(name)

    url = os.path.join(path, name, file).replace("\\", "/")  # if user is on windows
    file_path = os.path.join(name, file)

    if os.path.exists(file_path):
        log.warning(f"File already downloaded:\n{os.path.abspath(file_path)}")
        return
    else:
        log.info(
            f"Downloading {ouf.b(file, return_str=True)}\n{os.path.join(path,name,file)}"
        )

    if " " in url:
        url = quote(url, safe=":/")

    with tqdm(
        total=_get_file_size(url),
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        ncols=80,
    ) as pbar:
        urllib.request.urlretrieve(
            url,
            file_path,
            reporthook=lambda blocknum, bs, total: _update_progress(pbar, blocknum, bs),
        )


def _extract_html(url):
    """Extracts the html content of a webpage in "utf-8"

    Args:
        url(str): url to the location where all the data is stored.

    Returns:
        html_content(str): decoded html.
    """

    try:
        with urllib.request.urlopen(url) as response:
            html_content = response.read().decode(
                "utf-8"
            )  # Assuming the content is in UTF-8 encoding
    except urllib.error.URLError as e:
        log.warning(f"Failed to retrieve data from {url}. Error: {e}")

    return html_content


def _extract_names(name=None):
    """Extracts the names of the folders and files.

    Finds the names of either the folders if no name is given,
    or all the names of all files in the given folder.

    Args:
        name(str,optional): name of the folder from which the names should be extracted.

    Returns:
        list: If name is None, returns a list of all folders available.
              If name is not None, returns a list of all files available in the given 'name' folder.
    """

    url = "https://archive.compute.dtu.dk/files/public/projects/viscomp_data_repository"
    if name:
        datapath = os.path.join(url, name).replace("\\", "/")
        html_content = _extract_html(datapath)

        data_split = html_content.split(
            "files/public/projects/viscomp_data_repository/"
        )[3:]
        data_files = [
            element.split(" ")[0][(len(name) + 1) : -3] for element in data_split
        ]

        return data_files
    else:
        html_content = _extract_html(url)
        split = html_content.split('"icon-folder-open">')[2:]
        folders = [element.split(" ")[0][4:-4] for element in split]

        return folders



"""
Provides functionality for saving data from various file formats.


Example:
    ```python
    import qim3d
    
    vol = qim3d.examples.fly_150x256x256
    
    qim3d.io.save("fly.tif", vol)
    ```

    Volumes can also be saved with one file per slice:
    ```python
    import qim3d

    vol = qim3d.examples.fly_150x256x256
    
    qim3d.io.save("slices", vol, basename="fly-slices", sliced_dim=0)
    ```

"""

import datetime
import os

import dask.array as da
import h5py
import nibabel as nib
import numpy as np
import PIL
import tifffile
import zarr
from pydicom.dataset import FileDataset, FileMetaDataset
from pydicom.uid import UID

from qim3d.utils.logger import log
from qim3d.utils.misc import sizeof, stringify_path


class DataSaver:
    """Utility class for saving data to different file formats.

    Attributes:
        replace (bool): Specifies if an existing file with identical path is replaced.
        compression (bool): Specifies if the file is saved with Deflate compression (lossless).
        basename (str): Specifies the basename for a TIFF stack saved as several files
            (only relevant for TIFF stacks).
        sliced_dim (int): Specifies the dimension that is sliced in case a TIFF stack is saved
            as several files (only relevant for TIFF stacks)

    Methods:
        save_tiff(path,data): Save data to a TIFF file to the given path.
        load(path,data): Save data to the given path.

    Example:
        image = qim3d.examples.blobs_256x256
        saver = qim3d.io.DataSaver(compression=True)
        saver.save_tiff("image.tif",image)
    """

    def __init__(self, **kwargs):
        """Initializes a new instance of the DataSaver class.

        Args:
            replace (bool, optional): Specifies if an existing file with identical path should be replaced.
                Default is False.
            compression (bool, optional): Specifies if the file should be saved with Deflate compression.
                Default is False.
            basename (str, optional): Specifies the basename for a TIFF stack saved as several files
                (only relevant for TIFF stacks). Default is None
            sliced_dim (int, optional): Specifies the dimension that is sliced in case a TIFF stack is saved
                as several files (only relevant for TIFF stacks). Default is 0, i.e., the first dimension.
        """
        self.replace = kwargs.get("replace", False)
        self.compression = kwargs.get("compression", False)
        self.basename = kwargs.get("basename", None)
        self.sliced_dim = kwargs.get("sliced_dim", 0)

    def save_tiff(self, path, data):
        """Save data to a TIFF file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        tifffile.imwrite(path, data, compression=self.compression)

    def save_tiff_stack(self, path, data):
        """Save data as a TIFF stack containing slices in separate files to the given path.
        The slices will be named according to the basename plus a suffix with a zero-filled
        value corresponding to the slice number

        Args:
            path (str): The directory to save files to
            data (numpy.ndarray): The data to be saved
        """
        extension = ".tif"

        if data.ndim <= 2:
            path = os.path.join(path, self.basename, ".tif")
            self.save_tiff(path, data)
        else:
            # get number of total slices
            no_slices = data.shape[self.sliced_dim]
            # Get number of zero-fill values as the number of digits in the total number of slices
            zfill_val = len(str(no_slices))

            # Create index list
            idx = [slice(None)] * data.ndim

            # Iterate through each slice and save
            for i in range(no_slices):
                idx[self.sliced_dim] = i
                sliced = data[tuple(idx)]
                filename = self.basename + str(i).zfill(zfill_val) + extension
                filepath = os.path.join(path, filename)
                self.save_tiff(filepath, sliced)

            pattern_string = (
                filepath[: -(len(extension) + zfill_val)] + "-" * zfill_val + extension
            )

            log.info(
                f"Total of {no_slices} files saved following the pattern '{pattern_string}'"
            )

    def save_nifti(self, path, data):
        """Save data to a NIfTI file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        import nibabel as nib

        # Create header
        header = nib.Nifti1Header()
        header.set_data_dtype(data.dtype)

        # Create NIfTI image object
        img = nib.Nifti1Image(data, np.eye(4), header)

        # nib does automatically compress if filetype ends with .gz
        if self.compression and not path.endswith(".gz"):
            path += ".gz"
            log.warning("File extension '.gz' is added since compression is enabled.")

        if not self.compression and path.endswith(".gz"):
            path = path[:-3]
            log.warning(
                "File extension '.gz' is ignored since compression is disabled."
            )

        # Save image
        nib.save(img, path)

    def save_vol(self, path, data):
        """Save data to a VOL file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        # No support for compression yet
        if self.compression:
            raise NotImplementedError(
                "Saving compressed .vol files is not yet supported"
            )

        # Create custom .vgi metadata file
        metadata = ""
        metadata += "{volume1}\n"  # .vgi organization
        metadata += "[file1]\n"  # .vgi organization
        metadata += "Size = {} {} {}\n".format(
            data.shape[1], data.shape[2], data.shape[0]
        )  # Swap axes to match .vol format
        metadata += "Datatype = {}\n".format(str(data.dtype))  # Get datatype as string
        metadata += "Name = {}.vol\n".format(
            path.rsplit("/", 1)[-1][:-4]
        )  # Get filename without extension

        # Save metadata
        with open(path[:-4] + ".vgi", "w") as f:
            f.write(metadata)

        # Save data using numpy in binary format
        data.tofile(path[:-4] + ".vol")

    def save_h5(self, path, data):
        """Save data to a HDF5 file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        import h5py

        with h5py.File(path, "w") as f:
            f.create_dataset(
                "dataset", data=data, compression="gzip" if self.compression else None
            )

    def save_dicom(self, path, data):
        """Save data to a DICOM file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        import pydicom
        from pydicom.dataset import FileDataset, FileMetaDataset
        from pydicom.uid import UID

        # based on https://pydicom.github.io/pydicom/stable/auto_examples/input_output/plot_write_dicom.html

        # Populate required values for file meta information
        file_meta = FileMetaDataset()
        file_meta.MediaStorageSOPClassUID = UID("1.2.840.10008.5.1.4.1.1.2")
        file_meta.MediaStorageSOPInstanceUID = UID("1.2.3")
        file_meta.ImplementationClassUID = UID("1.2.3.4")

        # Create the FileDataset instance (initially no data elements, but file_meta
        # supplied)
        ds = FileDataset(path, {}, file_meta=file_meta, preamble=b"\0" * 128)

        ds.PatientName = "Test^Firstname"
        ds.PatientID = "123456"
        ds.StudyInstanceUID = "1.2.3.4.5"
        ds.SamplesPerPixel = 1
        ds.PixelRepresentation = 0
        ds.BitsStored = 16
        ds.BitsAllocated = 16
        ds.PhotometricInterpretation = "MONOCHROME2"
        ds.Rows = data.shape[1]
        ds.Columns = data.shape[2]
        ds.NumberOfFrames = data.shape[0]
        # Set the transfer syntax
        ds.is_little_endian = True
        ds.is_implicit_VR = True

        # Set creation date/time
        dt = datetime.datetime.now()
        ds.ContentDate = dt.strftime("%Y%m%d")
        timeStr = dt.strftime("%H%M%S.%f")  # long format with micro seconds
        ds.ContentTime = timeStr
        # Needs to be here because of bug in pydicom
        ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian

        # Reshape the data into a 1D array and convert to uint16
        data_1d = data.ravel().astype(np.uint16)
        # Convert the data to bytes
        data_bytes = data_1d.tobytes()
        # Add the data to the DICOM file
        ds.PixelData = data_bytes

        ds.save_as(path)

    def save_to_zarr(self, path, data):
        """ Saves a Dask array to a Zarr array on disk.

        Args:
            path (str): The path to the Zarr array on disk.
            data (dask.array): The Dask array to be saved to disk.

        Returns:
            zarr.core.Array: The Zarr array saved on disk.
        """
        assert isinstance(data, da.Array), 'data must be a dask array'

        # forces compute when saving to zarr
        da.to_zarr(data, path, compute=True, overwrite=self.replace, compressor=zarr.Blosc(cname='zstd', clevel=3, shuffle=2))
        
    
    def save_PIL(self, path, data):
        """Save data to a PIL file to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved
        """
        # No support for compression yet
        if self.compression and path.endswith(".png"):
            raise NotImplementedError("png does not support compression")
        elif not self.compression and path.endswith((".jpeg", ".jpg")):
            raise NotImplementedError("jpeg does not support no compression")

        # Convert to PIL image
        img = PIL.Image.fromarray(data)

        # Save image
        img.save(path)

    def save(self, path, data):
        """Save data to the given path.

        Args:
            path (str): The path to save file to
            data (numpy.ndarray): The data to be saved

        Raises:
            ValueError: If the provided path is an existing directory and self.basename is not provided
            ValueError: If the file format is not supported.
            ValueError: If the provided path does not exist and self.basename is not provided
            ValueError: If a file extension is not provided.
            ValueError: if a file with the specified path already exists and replace=False.

        Example:
            image = qim3d.examples.blobs_256x256
            saver = qim3d.io.DataSaver(compression=True)
            saver.save("image.tif",image)

        """

        path = stringify_path(path)
        isdir = os.path.isdir(path)
        _, ext = os.path.splitext(path)

        # If path is an existing directory
        if isdir:
            # If basename is provided
            if self.basename:
                # Save as tiff stack
                return self.save_tiff_stack(path, data)
            # If basename is not provided
            else:
                raise ValueError(
                    f"To save a stack as several TIFF files to the directory '{path}', please provide the keyword argument 'basename'. "
                    + "Otherwise, to save a single file, please provide a full path with a filename and valid extension."
                )

        # If path is not an existing directory
        else:
            # If there is no file extension in path and basename is provided
            if not ext and self.basename:
                # Make directory and save as tiff stack
                os.mkdir(path)
                log.info("Created directory '%s'!", path)
                return self.save_tiff_stack(path, data)

            # Check if a parent directory exists
            parentdir = os.path.dirname(path) or "."
            if os.path.isdir(parentdir):
                # If there is a file extension in the path
                if ext:
                    # If there is a basename
                    if self.basename:
                        # It will be unused and the user is informed accordingly
                        log.info("'basename' argument is unused")
                    # Check if a file with the given path already exists
                    if os.path.isfile(path) and not self.replace:
                        raise ValueError(
                            "A file with the provided path already exists. To replace it set 'replace=True'"
                        )

                    if path.endswith((".tif", ".tiff")):
                        return self.save_tiff(path, data)
                    elif path.endswith((".nii", "nii.gz")):
                        return self.save_nifti(path, data)
                    elif path.endswith(("TXRM", "XRM", "TXM")):
                        raise NotImplementedError(
                            "Saving TXRM files is not yet supported"
                        )
                    elif path.endswith((".h5")):
                        return self.save_h5(path, data)
                    elif path.endswith((".vol", ".vgi")):
                        return self.save_vol(path, data)
                    elif path.endswith((".dcm", ".DCM")):
                        return self.save_dicom(path, data)
                    elif path.endswith((".zarr")):
                        return self.save_to_zarr(path, data)
                    elif path.endswith((".jpeg",".jpg", ".png")):
                        return self.save_PIL(path, data)
                    else:
                        raise ValueError("Unsupported file format")
                # If there is no file extension in the path
                else:
                    raise ValueError(
                        "Please provide a file extension if you want to save as a single file."
                        + " Otherwise, please provide a basename to save as a TIFF stack"
                    )
            else:
                raise ValueError(
                    f"The directory '{parentdir}' does not exist.\n"
                    + "Please provide a valid directory or specify a basename if you want to save a tiff stack as several files to a folder that does not yet exist"
                )


def save(
    path, data, replace=False, compression=False, basename=None, sliced_dim=0, **kwargs
):
    """Save data to a specified file path.

    Args:
        path (str): The path to save file to
        data (numpy.ndarray): The data to be saved
        replace (bool, optional): Specifies if an existing file with identical path should be replaced.
            Default is False.
        compression (bool, optional): Specifies if the file should be saved with Deflate compression (lossless).
            Default is False.
        basename (str, optional): Specifies the basename for a TIFF stack saved as several files
            (only relevant for TIFF stacks). Default is None
        sliced_dim (int, optional): Specifies the dimension that is sliced in case a TIFF stack is saved
            as several files (only relevant for TIFF stacks). Default is 0, i.e., the first dimension.
        **kwargs: Additional keyword arguments to be passed to the DataSaver constructor

    Example:
        ```python
        import qim3d

        vol = qim3d.examples.blobs_256x256x256
        qim3d.io.save("blobs.tif", vol)
        ```
    """

    DataSaver(
        replace=replace,
        compression=compression,
        basename=basename,
        sliced_dim=sliced_dim,
        **kwargs,
    ).save(path, data)



from . import doi
from .progress_bar import ProgressBar
from .system import Memory

from .misc import (
    get_local_ip,
    port_from_str,
    gradio_header,
    sizeof,
    get_file_size,
    get_port_dict,
    get_css,
    downscale_img,
    scale_to_float16,
)
